"""
Test that the Handler constructor processes the supplied collection of service handlers
correctly.
"""

import pytest

from nexusrpc import HandlerError, LazyValue
from nexusrpc.handler import (
    CancelOperationContext,
    Handler,
    OperationHandler,
    StartOperationContext,
    StartOperationResultSync,
    operation_handler,
    service_handler,
)
from tests.helpers import DummySerializer, TestOperationTaskCancellation


def test_service_must_use_decorator():
    class Service:
        pass

    with pytest.raises(RuntimeError):
        _ = Handler([Service()])


def test_services_are_collected():
    class OpHandler(OperationHandler[int, int]):
        async def start(
            self,
            ctx: StartOperationContext,
            input: int,
        ) -> StartOperationResultSync[int]: ...

        async def cancel(
            self,
            ctx: CancelOperationContext,
            token: str,
        ) -> None: ...

    @service_handler
    class Service1:
        @operation_handler
        def op(self) -> OperationHandler[int, int]:
            return OpHandler()

    service_handlers = Handler([Service1()])
    assert service_handlers.service_handlers.keys() == {"Service1"}
    assert service_handlers.service_handlers["Service1"].service.name == "Service1"
    assert service_handlers.service_handlers["Service1"].operation_handlers.keys() == {
        "op"
    }


def test_service_names_must_be_unique():
    @service_handler(name="a")
    class Service1:
        pass

    @service_handler(name="a")
    class Service2:
        pass

    with pytest.raises(RuntimeError):
        _ = Handler([Service1(), Service2()])


@pytest.mark.asyncio
async def test_operations_must_have_decorator():
    @service_handler
    class TestService:
        async def op(self, _ctx: StartOperationContext, input: str) -> str:
            return input

    handler = Handler([TestService()])

    with pytest.raises(HandlerError, match="has no operation 'op'"):
        _ = await handler.start_operation(
            StartOperationContext(
                service=TestService.__name__,
                operation=TestService.op.__name__,
                headers={},
                request_id="test-req",
                task_cancellation=TestOperationTaskCancellation(),
                request_deadline=None,
                callback_url=None,
            ),
            LazyValue(
                serializer=DummySerializer(value="test"),
                headers={},
                stream=None,
            ),
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

import pytest

import nexusrpc
from nexusrpc._util import get_service_definition
from nexusrpc.handler import (
    OperationHandler,
    service_handler,
)
from nexusrpc.handler._core import ServiceHandler
from nexusrpc.handler._decorators import operation_handler


@dataclass()
class _DecoratorValidationTestCase:
    UserService: type[Any]
    UserServiceHandler: type[Any]
    expected_error_message_pattern: str


class MissingOperationFromDefinition(_DecoratorValidationTestCase):
    @nexusrpc.service
    class UserService:
        op_A: nexusrpc.Operation[int, str]
        op_B: nexusrpc.Operation[bool, float]

    class UserServiceHandler:
        @operation_handler
        def op_A(self) -> OperationHandler[int, str]: ...

    expected_error_message_pattern = (
        r"does not implement an operation with method name 'op_B'"
    )


class MethodNameDoesNotMatchDefinition(_DecoratorValidationTestCase):
    @nexusrpc.service
    class UserService:
        op_A: nexusrpc.Operation[int, str] = nexusrpc.Operation(name="foo")

    class UserServiceHandler:
        @operation_handler
        def op_A_incorrect_method_name(
            self,
        ) -> OperationHandler[int, str]: ...

    expected_error_message_pattern = (
        r"does not match an operation method name in the service definition."
    )


@pytest.mark.parametrize(
    "test_case",
    [
        MissingOperationFromDefinition,
        MethodNameDoesNotMatchDefinition,
    ],
)
def test_decorator_validates_definition_compliance(
    test_case: _DecoratorValidationTestCase,
):
    with pytest.raises(TypeError, match=test_case.expected_error_message_pattern):
        service_handler(service=test_case.UserService)(test_case.UserServiceHandler)


@dataclass()
class _ServiceHandlerInheritanceTestCase:
    UserServiceHandler: type[Any]
    expected_operations: set[str]


class ServiceHandlerInheritanceWithServiceDefinition(
    _ServiceHandlerInheritanceTestCase
):
    @nexusrpc.service
    class BaseUserService:
        base_op: nexusrpc.Operation[int, str]

    @nexusrpc.service
    class UserService:
        base_op: nexusrpc.Operation[int, str]
        child_op: nexusrpc.Operation[bool, float]

    @service_handler(service=BaseUserService)
    class BaseUserServiceHandler:
        @operation_handler
        def base_op(self) -> OperationHandler[int, str]: ...

    @service_handler(service=UserService)
    class UserServiceHandler(BaseUserServiceHandler):
        @operation_handler
        def child_op(self) -> OperationHandler[bool, float]: ...

    expected_operations = {"base_op", "child_op"}


class ServiceHandlerInheritanceWithoutDefinition(_ServiceHandlerInheritanceTestCase):
    @service_handler
    class BaseUserServiceHandler:
        @operation_handler
        def base_op_nc(self) -> OperationHandler[int, str]: ...

    @service_handler
    class UserServiceHandler(BaseUserServiceHandler):
        @operation_handler
        def child_op_nc(self) -> OperationHandler[bool, float]: ...

    expected_operations = {"base_op_nc", "child_op_nc"}


@pytest.mark.parametrize(
    "test_case",
    [
        ServiceHandlerInheritanceWithServiceDefinition,
        ServiceHandlerInheritanceWithoutDefinition,
    ],
)
def test_service_implementation_inheritance(
    test_case: _ServiceHandlerInheritanceTestCase,
):
    service_handler = ServiceHandler.from_user_instance(test_case.UserServiceHandler())

    assert set(service_handler.operation_handlers) == test_case.expected_operations
    assert (
        set(service_handler.service.operation_definitions)
        == test_case.expected_operations
    )


@dataclass()
class _ServiceDefinitionInheritanceTestCase:
    UserService: type[Any]
    expected_ops: set[str]


class ServiceDefinitionInheritance(_ServiceDefinitionInheritanceTestCase):
    @nexusrpc.service
    class BaseUserService:
        op_from_base_definition: nexusrpc.Operation[int, str]

    @nexusrpc.service
    class UserService(BaseUserService):
        op_from_child_definition: nexusrpc.Operation[bool, float]

    expected_ops = {
        "op_from_base_definition",
        "op_from_child_definition",
    }


@pytest.mark.parametrize(
    "test_case",
    [
        ServiceDefinitionInheritance,
    ],
)
def test_service_definition_inheritance_behavior(
    test_case: _ServiceDefinitionInheritanceTestCase,
):
    service_defn = get_service_definition(test_case.UserService)

    assert service_defn is not None, (
        f"{test_case.UserService.__name__} lacks __nexus_service__ attribute."
    )
    assert isinstance(service_defn, nexusrpc.ServiceDefinition), (
        "__nexus_service__ is not a nexusrpc.ServiceDefinition instance."
    )

    assert set(service_defn.operation_definitions) == test_case.expected_ops

    with pytest.raises(
        TypeError,
        match="does not implement an operation with method name 'op_from_child_definition'",
    ):

        @service_handler(service=test_case.UserService)
        class HandlerMissingChildOp:
            @operation_handler
            def op_from_base_definition(
                self,
            ) -> OperationHandler[int, str]: ...

        _ = HandlerMissingChildOp



"""
This module contains Handler classes. A Handler manages a collection of Nexus service
handlers. It receives and responds to incoming Nexus requests, dispatching to the
corresponding operation handler.
"""

# A description of the dispatch logic follows.

# There are two cases:

# Case 1: Every user service handler class has a corresponding service definition
# ===============================================================================

# I.e., there are service definitions that look like::

#     @service
#     class MyServiceDefinition:
#         my_op: nexusrpc.Operation[I, O]


# and every service handler class looks like::

#     @service_handler(service=MyServiceDefinition)
#     class MyServiceHandler:
#         @sync_operation
#         def my_op(self, ...)


# Import time
# -----------

# 1. The @service decorator builds a ServiceDefinition instance and attaches it to
#    MyServiceDefinition.

#    The ServiceDefinition contains `name` and a map of OperationDefinition instances, keyed
#    by OperationDefinition.name (this is the publicly advertised name).

#    OperationDefinition contains `name`, `method_name`, and input and output types, all
#    non-None.

#    Operation is the sugar version of OperationDefinition. It is identical to
#    OperationDefinition except that `method_name`, and the input and output types may be
#    None.

# 2. The @sync_operation decorator builds an Operation instance and attaches it to a factory
#    method that is attached to the my_op method object.

# 3. The @service_handler decorator acquires the ServiceDefinition instance from
#    MyServiceDefinition and attaches it to the MyServiceHandler class.


# Handler-registration time
# -------------------------

# 4. Handler.__init__ is called with [MyServiceHandler()]

# 5. A ServiceHandler instance is built from the user service handler class. This comprises
#    a ServiceDefinition and a map {op.name: OperationHandler}. The map is built by taking
#    every operation definition in the service definition and locating the operation handler
#    factory method whose *method name* matches the method name of the operation definition.

# 6. Finally we build a map {service_definition.name: ServiceHandler} using the service
#    definition in each ServiceHandler.

# Request-handling time
# ---------------------

# Now suppose a request has arrived for service s and operation o.

# 7. The Handler does self.service_handlers[s], yielding an instance of ServiceHandler.

# 8. The ServiceHandler does self.operation_handlers[o], yielding an instance of
#    OperationHandler

# Therefore we require that Handler.service_handlers and ServiceHandler.operation_handlers
# are keyed by the publicly advertised service and operation name respectively. This was
# achieved at steps (3) and (2) respectively.


# Case 2: There exists a user service handler class without a corresponding service definition
# ============================================================================================

# I.e., at least one user service handler class looks like::

#     @service_handler
#     class MyServiceHandler:
#         @sync_operation
#         def my_op(...)

# This follows Case 1 with the following differences at import time:

# - Step (1) does not occur.
# - At step (3) the ServiceDefinition is synthesized by the @service_handler decorator
#   from MyServiceHandler. An error will be thrown if the user did not supply input and
#   output type annotations on their operation handler methods (they do not have to do
#   this under Case 1).

from __future__ import annotations

import asyncio
import concurrent.futures
from abc import ABC, abstractmethod
from collections.abc import Awaitable, Mapping, Sequence
from dataclasses import dataclass
from typing import Any, Callable, Optional, Union, cast

from typing_extensions import Self, TypeGuard

from nexusrpc._common import HandlerError, HandlerErrorType
from nexusrpc._serializer import LazyValueT
from nexusrpc._service import ServiceDefinition
from nexusrpc._util import get_service_definition, is_async_callable

from ._common import (
    CancelOperationContext,
    OperationContext,
    StartOperationContext,
    StartOperationResultAsync,
    StartOperationResultSync,
)
from ._operation_handler import (
    MiddlewareSafeOperationHandler,
    OperationHandler,
    collect_operation_handler_factories_by_method_name,
)


class AbstractHandler(ABC):
    """
    .. warning::
        This API is experimental and may change in the future.
    """

    @abstractmethod
    def start_operation(
        self,
        ctx: StartOperationContext,
        input: LazyValueT,
    ) -> Union[
        StartOperationResultSync[Any],
        StartOperationResultAsync,
        Awaitable[
            Union[
                StartOperationResultSync[Any],
                StartOperationResultAsync,
            ]
        ],
    ]: ...

    @abstractmethod
    def cancel_operation(
        self, ctx: CancelOperationContext, token: str
    ) -> Union[None, Awaitable[None]]:
        """Handle a Cancel Operation request.

        Args:
            ctx: The operation context.
            token: The operation token.
        """
        ...


class BaseServiceCollectionHandler(AbstractHandler, ABC):
    """
    A Nexus handler, managing a collection of Nexus service handlers.

    Operation requests are delegated to a :py:class:`ServiceHandler` based on the service
    name in the operation context.

    .. warning::
        This API is experimental and may change in the future.
    """

    def __init__(
        self,
        user_service_handlers: Sequence[Any],
        executor: Optional[concurrent.futures.Executor] = None,
    ):
        """Initialize a :py:class:`Handler` instance from user service handler instances.

        The user service handler instances must have been decorated with the
        :py:func:`@nexusrpc.handler.service_handler` decorator.

        Args:
            user_service_handlers: A sequence of user service handlers.
            executor: A concurrent.futures.Executor in which to run non-`async def` operation handlers.
        """
        self.executor = _Executor(executor) if executor else None
        self.service_handlers = self._register_service_handlers(user_service_handlers)

    def _register_service_handlers(
        self, user_service_handlers: Sequence[Any]
    ) -> Mapping[str, ServiceHandler]:
        service_handlers = {}
        for sh in user_service_handlers:
            if isinstance(sh, type):
                raise TypeError(
                    f"Expected a service instance, but got a class: {type(sh)}. "
                    "Nexus service handlers must be supplied as instances, not classes."
                )
            # Users may register ServiceHandler instances directly.
            if not isinstance(sh, ServiceHandler):
                # It must be a user service handler instance (i.e. an instance of a class
                # decorated with @nexusrpc.handler.service_handler).
                sh = ServiceHandler.from_user_instance(sh)
            if sh.service.name in service_handlers:
                raise RuntimeError(
                    f"Service '{sh.service.name}' has already been registered."
                )
            service_handlers[sh.service.name] = sh
        return service_handlers

    def _get_service_handler(self, service_name: str) -> ServiceHandler:
        """Return a service handler, given the service name."""
        service = self.service_handlers.get(service_name)
        if service is None:
            raise HandlerError(
                f"No handler for service '{service_name}'.",
                type=HandlerErrorType.NOT_FOUND,
            )
        return service


class Handler(BaseServiceCollectionHandler):
    """
    A Nexus handler manages a collection of Nexus service handlers.

    Operation requests are dispatched to a :py:class:`ServiceHandler` based on the
    service name in the operation context.

    This class supports user operation handlers that are either `async def` or `def`. If
    `def` user operation handlers are to be supported, an executor must be provided.

    The methods of this class itself are `async def`. There is currently no alternative
    Handler class with `def` methods.

    .. warning::
        This API is experimental and may change in the future.

    Example:
        .. code-block:: python

            import concurrent.futures
            from nexusrpc.handler import Handler

            # Create service handler instances
            my_service = MyServiceHandler()

            # Create handler with async operations only
            handler = Handler([my_service])

            # Create handler that supports both async and sync operations
            executor = concurrent.futures.ThreadPoolExecutor(max_workers=10)
            handler = Handler([my_service], executor=executor)

            # Use handler to process requests
            result = await handler.start_operation(ctx, input_lazy_value)
    """

    def __init__(
        self,
        user_service_handlers: Sequence[Any],
        executor: Optional[concurrent.futures.Executor] = None,
        middleware: Sequence[OperationHandlerMiddleware] | None = None,
    ):
        self._middleware = cast(Sequence[OperationHandlerMiddleware], middleware or [])
        super().__init__(user_service_handlers, executor=executor)
        if not self.executor:
            self._validate_all_operation_handlers_are_async()

    async def start_operation(
        self,
        ctx: StartOperationContext,
        input: LazyValueT,
    ) -> Union[
        StartOperationResultSync[Any],
        StartOperationResultAsync,
    ]:
        """Handle a Start Operation request.

        Args:
            ctx: The operation context.
            input: The input to the operation, as a LazyValue.
        """
        service_handler = self._get_service_handler(ctx.service)
        op_handler = self._get_operation_handler(ctx, service_handler, ctx.operation)

        op_defn = service_handler.service.operation_definitions[ctx.operation]
        deserialized_input = await input.consume(as_type=op_defn.input_type)
        return await op_handler.start(ctx, deserialized_input)

    async def cancel_operation(self, ctx: CancelOperationContext, token: str) -> None:
        """Handle a Cancel Operation request.

        Args:
            ctx: The operation context.
            token: The operation token.
        """
        service_handler = self._get_service_handler(ctx.service)
        op_handler = self._get_operation_handler(ctx, service_handler, ctx.operation)
        return await op_handler.cancel(ctx, token)

    def _get_operation_handler(
        self, ctx: OperationContext, service_handler: ServiceHandler, operation: str
    ) -> MiddlewareSafeOperationHandler:
        """
        Get the specified handler for the specified operation from the given service_handler and apply all middleware.
        """
        op_handler: MiddlewareSafeOperationHandler = _EnsuredAwaitableOperationHandler(
            self.executor, service_handler.get_operation_handler(operation)
        )

        for middleware in reversed(self._middleware):
            op_handler = middleware.intercept(ctx, op_handler)

        return op_handler

    def _validate_all_operation_handlers_are_async(self) -> None:
        for service_handler in self.service_handlers.values():
            for op_handler in service_handler.operation_handlers.values():
                _ = self._assert_async_callable(op_handler.start)
                _ = self._assert_async_callable(op_handler.cancel)

    def _assert_async_callable(
        self, method: Callable[..., Any]
    ) -> TypeGuard[Callable[..., Awaitable[Any]]]:
        if not is_async_callable(method):
            raise RuntimeError(
                f"Operation handler method {method} is not an `async def` method, "
                f"but you have not supplied an executor to nexusrpc.handler.Handler. "
            )
        return True


@dataclass(frozen=True)
class ServiceHandler:
    """Internal representation of an instance of a user's service handler class.

    A service handler is a class decorated with
    :py:func:`@nexusrpc.handler.service_handler` that defines operation handler methods
    using decorators such as :py:func:`@nexusrpc.handler.operation_handler`.

    Instances of this class are created automatically from user service implementation
    instances on creation of a Handler instance, at Nexus handler start time. While the
    user's class defines operation handlers as factory methods to be called at handler
    start time, this class contains the :py:class:`OperationHandler` instances
    themselves.

    You may create instances of this class manually and pass them to the Handler
    constructor, for example when programmatically creating Nexus service
    implementations.
    """

    service: ServiceDefinition
    operation_handlers: dict[str, OperationHandler[Any, Any]]

    @classmethod
    def from_user_instance(cls, user_instance: Any) -> Self:
        """Create a :py:class:`ServiceHandler` from a user service instance."""

        service = get_service_definition(user_instance.__class__)
        if not isinstance(service, ServiceDefinition):
            raise RuntimeError(
                f"Service '{user_instance}' does not have a service definition. "
                f"Use the :py:func:`@nexusrpc.handler.service_handler` decorator on your class to define "
                f"a Nexus service implementation."
            )

        # Construct a map of operation handlers keyed by the op name from the service
        # definition (i.e. by the name by which the operation can be requested)
        factories_by_method_name = collect_operation_handler_factories_by_method_name(
            user_instance.__class__, service
        )
        op_handlers = {
            name: factories_by_method_name[defn.method_name](user_instance)
            for name, defn in service.operation_definitions.items()
        }
        return cls(
            service=service,
            operation_handlers=op_handlers,
        )

    def get_operation_handler(self, operation_name: str) -> OperationHandler[Any, Any]:
        """Return an operation handler, given the operation name."""
        if operation_name not in self.service.operation_definitions:
            raise HandlerError(
                f"Nexus service definition '{self.service.name}' has no operation "
                f"'{operation_name}'. There are {len(self.service.operation_definitions)} operations "
                f"in the definition.",
                type=HandlerErrorType.NOT_FOUND,
            )
        operation_handler = self.operation_handlers.get(operation_name)
        if operation_handler is None:
            raise HandlerError(
                f"Nexus service implementation '{self.service.name}' has no handler for "
                f"operation '{operation_name}'. There are {len(self.operation_handlers)} "
                f"available operation handlers.",
                type=HandlerErrorType.NOT_FOUND,
            )
        return operation_handler


class _Executor:
    """An executor for synchronous functions."""

    # Users are require to pass in a `concurrent.futures.Executor` in order to use
    # non-`async def`s. This is what `run_in_executor` is documented to require.
    # This means that nexusrpc initially has a hard-coded dependency on the asyncio
    # event loop; if necessary we can add the ability to pass in an event loop
    # implementation at the level of Handler.

    def __init__(self, executor: concurrent.futures.Executor):
        self._executor = executor

    def submit_to_event_loop(
        self, fn: Callable[..., Any], *args: Any
    ) -> Awaitable[Any]:
        return asyncio.get_event_loop().run_in_executor(self._executor, fn, *args)

    def submit(
        self, fn: Callable[..., Any], *args: Any
    ) -> concurrent.futures.Future[Any]:
        return self._executor.submit(fn, *args)


class OperationHandlerMiddleware(ABC):
    """
    Middleware for operation handlers.

    This should be extended by any operation handler middelware.

    .. warning::
        This API is experimental and may change in the future.
    """

    @abstractmethod
    def intercept(
        self,
        ctx: OperationContext,  # type: ignore[reportUnusedParameter]
        next: MiddlewareSafeOperationHandler,
    ) -> MiddlewareSafeOperationHandler:
        """
        Method called for intercepting operation handlers.

        Args:
            ctx: The :py:class:`OperationContext` that will be passed to the operation handler.
            next: The underlying operation handler that this middleware
                should delegate to.

        Returns:
            The new middleware that will be used to invoke
            :py:attr:`OperationHandler.start` or :py:attr:`OperationHandler.cancel`.
        """
        ...


class _EnsuredAwaitableOperationHandler(MiddlewareSafeOperationHandler):
    """
    An :py:class:`AwaitableOperationHandler` that wraps an :py:class:`OperationHandler` and uses an :py:class:`_Executor` to ensure
    that the :py:attr:`start` and :py:attr:`cancel` methods are awaitable.
    """

    def __init__(
        self,
        executor: _Executor | None,
        op_handler: OperationHandler[Any, Any],
    ):
        self._executor = executor
        self._op_handler = op_handler

    async def start(
        self, ctx: StartOperationContext, input: Any
    ) -> StartOperationResultSync[Any] | StartOperationResultAsync:
        """
        Start the operation using the wrapped :py:class:`OperationHandler`.
        """
        if is_async_callable(self._op_handler.start):
            return await self._op_handler.start(ctx, input)
        else:
            assert self._executor
            return await self._executor.submit_to_event_loop(
                self._op_handler.start, ctx, input
            )

    async def cancel(self, ctx: CancelOperationContext, token: str) -> None:
        """
        Cancel an operation using the wrapped :py:class:`OperationHandler`.
        """
        if is_async_callable(self._op_handler.cancel):
            return await self._op_handler.cancel(ctx, token)
        else:
            assert self._executor
            return self._executor.submit(self._op_handler.cancel, ctx, token).result()



from __future__ import annotations

import typing
import warnings
from collections.abc import Awaitable
from typing import Any, Callable, Optional, TypeVar, Union, overload

import nexusrpc.handler._syncio
from nexusrpc._common import InputT, OutputT, ServiceHandlerT
from nexusrpc._service import Operation
from nexusrpc._util import (
    get_callable_name,
    get_service_definition,
    is_async_callable,
    set_operation,
    set_operation_factory,
    set_service_definition,
)
from nexusrpc.handler._common import StartOperationContext
from nexusrpc.handler._util import (
    get_start_method_input_and_output_type_annotations,
)

from ._operation_handler import (
    OperationHandler,
    SyncOperationHandler,
    collect_operation_handler_factories_by_method_name,
    service_definition_from_operation_handler_methods,
    validate_operation_handler_methods,
)


@overload
def service_handler(cls: type[ServiceHandlerT]) -> type[ServiceHandlerT]: ...


@overload
def service_handler(
    *,
    service: Optional[type[Any]] = None,
) -> Callable[[type[ServiceHandlerT]], type[ServiceHandlerT]]: ...


@overload
def service_handler(
    *, name: str
) -> Callable[[type[ServiceHandlerT]], type[ServiceHandlerT]]: ...


def service_handler(
    cls: Optional[type[ServiceHandlerT]] = None,
    *,
    service: Optional[type[Any]] = None,
    name: Optional[str] = None,
) -> Union[
    type[ServiceHandlerT], Callable[[type[ServiceHandlerT]], type[ServiceHandlerT]]
]:
    """Decorator that marks a class as a Nexus service handler.

    A service handler is a class that implements the Nexus service by providing
    operation handler implementations for all operations in the service.

    The class should implement Nexus operation handlers as methods decorated with
    operation handler decorators such as :py:func:`@nexusrpc.handler.operation_handler`.

    Args:
        cls: The service handler class to decorate.
        service: The service definition that the service handler implements.
        name: Optional name to use for the service, if a service definition is not provided.
              `service` and `name` are mutually exclusive. If neither is provided, the
              class name will be used.

    Example:
        .. code-block:: python

            from nexusrpc.handler import service_handler, sync_operation

            @service_handler(service=MyService)
            class MyServiceHandler:
                @sync_operation
                async def my_operation(
                    self, ctx: StartOperationContext, input: MyInput
                ) -> MyOutput:
                    return MyOutput(processed=input.data)
    """
    if service and name:
        raise ValueError(
            "You cannot specify both service and name: "
            "if you provide a service then the name will be taken from the service."
        )
    _service = None
    if service:
        _service = get_service_definition(service)
        if not _service:
            raise ValueError(
                f"{service} is not a valid Nexus service definition. "
                f"Use the @nexusrpc.service decorator on a class to define a Nexus service definition."
            )

    def decorator(cls: type[ServiceHandlerT]) -> type[ServiceHandlerT]:
        # The name by which the service must be addressed in Nexus requests.
        _name = (
            _service.name if _service else name if name is not None else cls.__name__
        )
        if not _name:
            raise ValueError("Service name must not be empty.")
        factories_by_method_name = collect_operation_handler_factories_by_method_name(
            cls, _service
        )
        service = _service or service_definition_from_operation_handler_methods(
            _name, factories_by_method_name
        )
        validate_operation_handler_methods(cls, factories_by_method_name, service)
        set_service_definition(cls, service)
        return cls

    if cls is None:
        return decorator

    return decorator(cls)


OperationHandlerFactoryT = TypeVar(
    "OperationHandlerFactoryT", bound=Callable[[Any], OperationHandler[Any, Any]]
)


@overload
def operation_handler(
    method: OperationHandlerFactoryT,
) -> OperationHandlerFactoryT: ...


@overload
def operation_handler(
    *, name: Optional[str] = None
) -> Callable[[OperationHandlerFactoryT], OperationHandlerFactoryT]: ...


def operation_handler(
    method: Optional[OperationHandlerFactoryT] = None,
    *,
    name: Optional[str] = None,
) -> Union[
    OperationHandlerFactoryT,
    Callable[[OperationHandlerFactoryT], OperationHandlerFactoryT],
]:
    """
    Decorator marking an operation handler factory method in a service handler class.

    An operation handler factory method is a method that takes no arguments other than
    `self` and returns an :py:class:`OperationHandler` instance.

    Args:
        method: The method to decorate.
        name: Optional name for the operation. If not provided, the method name will be used.
    """

    def decorator(
        method: OperationHandlerFactoryT,
    ) -> OperationHandlerFactoryT:
        # Extract input and output types from the return type annotation
        input_type = None
        output_type = None

        return_type = typing.get_type_hints(method).get("return")
        if typing.get_origin(return_type) == OperationHandler:
            type_args = typing.get_args(return_type)
            if len(type_args) == 2:
                input_type, output_type = type_args
            else:
                warnings.warn(
                    f"OperationHandler return type should have two type parameters (input and output type), "
                    f"but operation {method.__name__} has {len(type_args)} type parameters: {type_args}"
                )

        op: Operation[Any, Any] = Operation(
            name=name or method.__name__,
            input_type=input_type,
            output_type=output_type,
        )
        op.method_name = method.__name__
        set_operation(method, op)
        return method

    if method is None:
        return decorator

    return decorator(method)


F = TypeVar("F", bound=Callable[[Any, StartOperationContext, Any], Any])


@overload
def sync_operation(
    start: Callable[
        [ServiceHandlerT, StartOperationContext, InputT], Awaitable[OutputT]
    ],
) -> Callable[[ServiceHandlerT, StartOperationContext, InputT], Awaitable[OutputT]]: ...


@overload
def sync_operation(
    start: Callable[[ServiceHandlerT, StartOperationContext, InputT], OutputT],
) -> Callable[[ServiceHandlerT, StartOperationContext, InputT], OutputT]: ...


@overload
def sync_operation(
    *,
    name: Optional[str] = None,
) -> Callable[
    [Callable[[ServiceHandlerT, StartOperationContext, InputT], Any]],
    Callable[[ServiceHandlerT, StartOperationContext, InputT], Any],
]: ...


def sync_operation(
    start: Optional[F] = None,
    *,
    name: Optional[str] = None,
) -> Union[F, Callable[[F], F]]:
    """
    Decorator marking a method as the start method for a synchronous operation.

    Example:
        .. code-block:: python

            import httpx
            from nexusrpc.handler import service_handler, sync_operation

            @service_handler
            class MyServiceHandler:
                @sync_operation
                async def process_data(
                    self, ctx: StartOperationContext, input: str
                ) -> str:
                    # You can use asynchronous I/O libraries
                    async with httpx.AsyncClient() as client:
                        response = await client.get("https://api.example.com/data")

                    data = response.json()
                    return f"Processed: {data}"
    """

    def decorator(start: F) -> F:
        def operation_handler_factory(
            self: Any,
        ) -> OperationHandler[Any, Any]:
            if is_async_callable(start):

                async def asyncio_start(ctx: StartOperationContext, input: Any) -> Any:
                    return await start(self, ctx, input)

                asyncio_start.__doc__ = start.__doc__
                return SyncOperationHandler(asyncio_start)
            else:

                def _start(ctx: StartOperationContext, input: Any) -> Any:
                    return start(self, ctx, input)

                _start.__doc__ = start.__doc__
                return nexusrpc.handler._syncio.SyncOperationHandler(_start)

        # Type inspection happens here at @sync_operation decoration time, before we know
        # if a service definition will be provided to @service_handler. While we could
        # theoretically defer this inspection until @service_handler time (and skip it when
        # a service definition provides the types), doing so would require storing the
        # original method reference and implementing lazy evaluation. The added complexity
        # isn't worth the marginal performance benefit. When types are None and a service
        # definition is provided, type validation is simply skipped (see
        # validate_operation_handler_methods in _operation_handler.py).
        input_type, output_type = get_start_method_input_and_output_type_annotations(  # type: ignore[var-annotated]
            start  # type: ignore[arg-type]
        )

        method_name = get_callable_name(start)
        op = Operation(
            name=name or method_name,
            input_type=input_type,
            output_type=output_type,
        )
        op.method_name = method_name
        set_operation(operation_handler_factory, op)

        set_operation_factory(start, operation_handler_factory)
        return start

    if start is None:
        return decorator

    return decorator(start)



"""
Test that operation decorators result in operation definitions with the correct name and
input/ouput types.
"""

from dataclasses import dataclass
from typing import Any, Optional

import pytest

import nexusrpc
from nexusrpc._util import get_operation, get_service_definition
from nexusrpc.handler import (
    OperationHandler,
    service_handler,
)
from nexusrpc.handler._decorators import operation_handler


@dataclass
class Input:
    pass


@dataclass
class Output:
    pass


@dataclass
class _TestCase:
    Service: type[Any]
    expected_operations: dict[str, nexusrpc.Operation[Any, Any]]
    Contract: Optional[type[Any]] = None


class ManualOperationHandler(_TestCase):
    @service_handler
    class Service:
        @operation_handler
        def operation(self) -> OperationHandler[Input, Output]: ...

    expected_operations = {
        "operation": nexusrpc.Operation(
            name="operation",
            input_type=Input,
            output_type=Output,
        ),
    }


class ManualOperationHandlerWithNameOverride(_TestCase):
    @service_handler
    class Service:
        @operation_handler(name="operation-name")
        def operation(self) -> OperationHandler[Input, Output]: ...

    expected_operations = {
        "operation": nexusrpc.Operation(
            name="operation-name",
            input_type=Input,
            output_type=Output,
        ),
    }


class SyncOperation(_TestCase):
    @service_handler
    class Service:
        @operation_handler
        def sync_operation_handler(
            self,
        ) -> OperationHandler[Input, Output]: ...

    expected_operations = {
        "sync_operation_handler": nexusrpc.Operation(
            name="sync_operation_handler",
            input_type=Input,
            output_type=Output,
        ),
    }


class SyncOperationWithOperationHandlerNameOverride(_TestCase):
    @service_handler
    class Service:
        @operation_handler(name="sync-operation-name")
        def sync_operation_handler(
            self,
        ) -> OperationHandler[Input, Output]: ...

    expected_operations = {
        "sync_operation_handler": nexusrpc.Operation(
            name="sync-operation-name",
            input_type=Input,
            output_type=Output,
        ),
    }


class ManualOperationWithContract(_TestCase):
    @nexusrpc.service
    class Contract:
        operation: nexusrpc.Operation[Input, Output]

    @service_handler(service=Contract)
    class Service:
        @operation_handler
        def operation(self) -> OperationHandler[Input, Output]: ...

    expected_operations = {
        "operation": nexusrpc.Operation(
            name="operation",
            input_type=Input,
            output_type=Output,
        ),
    }


class ManualOperationWithContractNameOverrideAndOperationHandlerNameOverride(_TestCase):
    @nexusrpc.service
    class Contract:
        operation: nexusrpc.Operation[Input, Output] = nexusrpc.Operation(
            name="operation-override",
        )

    @service_handler(service=Contract)
    class Service:
        @operation_handler(name="operation-override")
        def operation(self) -> OperationHandler[Input, Output]: ...

    expected_operations = {
        "operation": nexusrpc.Operation(
            name="operation-override",
            input_type=Input,
            output_type=Output,
        ),
    }


@pytest.mark.parametrize(
    "test_case",
    [
        ManualOperationHandler,
        ManualOperationHandlerWithNameOverride,
        SyncOperation,
        SyncOperationWithOperationHandlerNameOverride,
        ManualOperationWithContract,
        ManualOperationWithContractNameOverrideAndOperationHandlerNameOverride,
    ],
)
@pytest.mark.asyncio
async def test_collected_operation_definitions(
    test_case: type[_TestCase],
):
    service = get_service_definition(test_case.Service)
    assert isinstance(service, nexusrpc.ServiceDefinition)
    if test_case.Contract:
        defn = get_service_definition(test_case.Contract)
        assert isinstance(defn, nexusrpc.ServiceDefinition)
        assert defn.name == service.name
    else:
        assert service.name == "Service"

    for method_name, expected_op in test_case.expected_operations.items():
        actual_op = get_operation(getattr(test_case.Service, method_name))
        assert isinstance(actual_op, nexusrpc.Operation)
        assert actual_op.name == expected_op.name
        assert actual_op.input_type == expected_op.input_type
        assert actual_op.output_type == expected_op.output_type



import concurrent.futures
import logging
import uuid
from typing import Any

import pytest

from nexusrpc import LazyValue
from nexusrpc.handler import (
    CancelOperationContext,
    Handler,
    MiddlewareSafeOperationHandler,
    OperationContext,
    OperationHandler,
    OperationHandlerMiddleware,
    StartOperationContext,
    StartOperationResultAsync,
    StartOperationResultSync,
    operation_handler,
    service_handler,
    sync_operation,
)
from tests.helpers import DummySerializer, TestOperationTaskCancellation

_operation_results: dict[str, int] = {}

logger = logging.getLogger()


class MyAsyncOperationHandler(OperationHandler[int, int]):
    async def start(
        self, ctx: StartOperationContext, input: int
    ) -> StartOperationResultAsync:
        token = str(uuid.uuid4())
        _operation_results[token] = input + 1
        return StartOperationResultAsync(token)

    async def cancel(self, ctx: CancelOperationContext, token: str) -> None:
        del _operation_results[token]


@service_handler
class MyService:
    @operation_handler
    def incr(self) -> OperationHandler[int, int]:
        return MyAsyncOperationHandler()


@service_handler
class MyServiceSync:
    @sync_operation
    def incr(self, ctx: StartOperationContext, input: int) -> int:  # type: ignore[reportUnusedParameter]
        return input + 1


class CountingMiddleware(OperationHandlerMiddleware):
    def __init__(self) -> None:
        self.num_start = 0
        self.num_cancel = 0

    def intercept(
        self, ctx: OperationContext, next: MiddlewareSafeOperationHandler
    ) -> MiddlewareSafeOperationHandler:
        return CountingOperationHandler(next, self)


class CountingOperationHandler(MiddlewareSafeOperationHandler):
    """
    An :py:class:`AwaitableOperationHandler` that wraps a counting middleware
    that counts the number of calls to each handler method.
    """

    def __init__(
        self,
        next: MiddlewareSafeOperationHandler,
        middleware: CountingMiddleware,
    ) -> None:
        self._next = next
        self._middleware = middleware

    async def start(
        self, ctx: StartOperationContext, input: Any
    ) -> StartOperationResultSync[Any] | StartOperationResultAsync:
        self._middleware.num_start += 1
        return await self._next.start(ctx, input)

    async def cancel(self, ctx: CancelOperationContext, token: str) -> None:
        self._middleware.num_cancel += 1
        return await self._next.cancel(ctx, token)


class MustBeFirstMiddleware(OperationHandlerMiddleware):
    def __init__(self, counter: CountingMiddleware) -> None:
        self._counter = counter

    def intercept(
        self, ctx: OperationContext, next: MiddlewareSafeOperationHandler
    ) -> MiddlewareSafeOperationHandler:
        return MustBeFirstOperationHandler(next, self._counter)


class MustBeFirstOperationHandler(MiddlewareSafeOperationHandler):
    """
    An :py:class:`AwaitableOperationHandler` that wraps a counting middleware
    and asserts that the wrapped middleware has a count of 0 for each handler method
    """

    def __init__(
        self,
        next: MiddlewareSafeOperationHandler,
        counter: CountingMiddleware,
    ) -> None:
        self._next = next
        self._counter = counter

    async def start(
        self, ctx: StartOperationContext, input: Any
    ) -> StartOperationResultSync[Any] | StartOperationResultAsync:
        assert self._counter.num_start == 0
        logger.info("%s.%s: start operation", ctx.service, ctx.operation)

        result = await self._next.start(ctx, input)

        if isinstance(result, StartOperationResultAsync):
            logger.info(
                "%s.%s: start operation completed async. token=%s",
                ctx.service,
                ctx.operation,
                result.token,
            )
        else:
            logger.info(
                "%s.%s: start operation completed sync. value=%s",
                ctx.service,
                ctx.operation,
                result.value,
            )

        return result

    async def cancel(self, ctx: CancelOperationContext, token: str) -> None:
        assert self._counter.num_cancel == 0
        logger.info("%s.%s: cancel token=%s", ctx.service, ctx.operation, token)
        return await self._next.cancel(ctx, token)


@pytest.mark.asyncio
async def test_async_operation_middleware_applied():
    counting_middleware = CountingMiddleware()
    handler = Handler(
        user_service_handlers=[MyService()],
        middleware=[
            MustBeFirstMiddleware(counting_middleware),
            counting_middleware,
        ],
    )
    start_ctx = StartOperationContext(
        service="MyService",
        operation="incr",
        headers={},
        request_id="request_id",
        task_cancellation=TestOperationTaskCancellation(),
    )
    start_result = await handler.start_operation(
        start_ctx, LazyValue(DummySerializer(1), headers={})
    )
    assert isinstance(start_result, StartOperationResultAsync)
    assert start_result.token

    cancel_ctx = CancelOperationContext(
        service="MyService",
        operation="incr",
        headers={},
        task_cancellation=TestOperationTaskCancellation(),
    )
    await handler.cancel_operation(cancel_ctx, start_result.token)
    assert start_result.token not in _operation_results

    assert counting_middleware.num_start == 1
    assert counting_middleware.num_cancel == 1


@pytest.mark.asyncio
async def test_sync_operation_middleware_applied():
    counting_middleware = CountingMiddleware()
    handler = Handler(
        user_service_handlers=[MyServiceSync()],
        executor=concurrent.futures.ThreadPoolExecutor(),
        middleware=[
            MustBeFirstMiddleware(counting_middleware),
            counting_middleware,
        ],
    )
    start_ctx = StartOperationContext(
        service="MyServiceSync",
        operation="incr",
        headers={},
        request_id="request_id",
        task_cancellation=TestOperationTaskCancellation(),
    )
    start_result = await handler.start_operation(
        start_ctx, LazyValue(DummySerializer(1), headers={})
    )
    assert isinstance(start_result, StartOperationResultSync)
    assert start_result.value == 2

    assert counting_middleware.num_start == 1
    assert counting_middleware.num_cancel == 0



"""
Test runtime behavior of operation handlers invoked through Handler.start_operation().

This file tests actual execution behavior, distinct from:
- Decoration-time validation (test_service_handler_decorator_validates_against_service_contract.py)
- Handler constructor validation (test_handler_validates_service_handler_collection.py)
"""

import pytest

from nexusrpc import LazyValue, Operation, service
from nexusrpc.handler import (
    CancelOperationContext,
    Handler,
    OperationHandler,
    StartOperationContext,
    StartOperationResultSync,
    operation_handler,
    service_handler,
)
from nexusrpc.handler._decorators import sync_operation
from tests.helpers import DummySerializer, TestOperationTaskCancellation


@pytest.mark.asyncio
async def test_handler_can_return_covariant_type():
    class Superclass:
        pass

    class Subclass(Superclass):
        pass

    @service
    class CovariantService:
        op_handler: Operation[None, Superclass]
        inline: Operation[None, Superclass]

    class ValidOperationHandler(OperationHandler[None, Superclass]):
        async def start(
            self, ctx: StartOperationContext, input: None
        ) -> StartOperationResultSync[Subclass]:
            return StartOperationResultSync(Subclass())

        async def cancel(self, ctx: CancelOperationContext, token: str) -> None:
            pass

    @service_handler(service=CovariantService)
    class CovariantServiceHandler:
        @operation_handler
        def op_handler(self) -> OperationHandler[None, Superclass]:
            return ValidOperationHandler()

        @sync_operation
        async def inline(self, ctx: StartOperationContext, input: None) -> Superclass:  # pyright: ignore[reportUnusedParameter]
            return Subclass()

    handler = Handler([CovariantServiceHandler()])

    result = await handler.start_operation(
        StartOperationContext(
            service=CovariantService.__name__,
            operation=CovariantService.op_handler.name,
            headers={},
            request_id="test-req",
            task_cancellation=TestOperationTaskCancellation(),
            request_deadline=None,
            callback_url=None,
        ),
        LazyValue(
            serializer=DummySerializer(None),
            headers={},
            stream=None,
        ),
    )
    assert type(result) is StartOperationResultSync
    assert type(result.value) is Subclass

    result = await handler.start_operation(
        StartOperationContext(
            service=CovariantService.__name__,
            operation=CovariantService.inline.name,
            headers={},
            request_id="test-req",
            task_cancellation=TestOperationTaskCancellation(),
            request_deadline=None,
            callback_url=None,
        ),
        LazyValue(
            serializer=DummySerializer(None),
            headers={},
            stream=None,
        ),
    )
    assert type(result) is StartOperationResultSync
    assert type(result.value) is Subclass



from dataclasses import dataclass

import pytest

import nexusrpc


class Output:
    pass


@dataclass()
class _TestCase:
    Contract: type
    expected_error: Exception


class DuplicateOperationNameOverride(_TestCase):
    class Contract:
        a: nexusrpc.Operation[None, Output] = nexusrpc.Operation(name="a")
        b: nexusrpc.Operation[int, str] = nexusrpc.Operation(name="a")

    expected_error = RuntimeError(
        r"Operation 'a' in service .* is defined multiple times"
    )


@pytest.mark.parametrize(
    "test_case",
    [
        DuplicateOperationNameOverride,
    ],
)
def test_operation_validation(
    test_case: type[_TestCase],
):
    with pytest.raises(
        type(test_case.expected_error),
        match=str(test_case.expected_error),
    ):
        nexusrpc.service(test_case.Contract)



from dataclasses import dataclass
from typing import Any

import pytest

import nexusrpc
from nexusrpc.handler import (
    StartOperationContext,
    service_handler,
    sync_operation,
)


@dataclass()
class _InterfaceImplementationTestCase:
    Interface: type
    Impl: type
    error_message: str | None


class _InvalidInputTestCase(_InterfaceImplementationTestCase):
    error_message = "OperationHandler input type mismatch"


class _InvalidOutputTestCase(_InterfaceImplementationTestCase):
    error_message = "OperationHandler output type mismatch"


class _ValidTestCase(_InterfaceImplementationTestCase):
    error_message = None


class ValidImpl(_ValidTestCase):
    @nexusrpc.service
    class Interface:
        op: nexusrpc.Operation[None, None]

        def unrelated_method(self) -> None: ...

    class Impl:
        @sync_operation
        async def op(self, _ctx: StartOperationContext, _input: None) -> None: ...


class ValidImplWithEmptyInterfaceAndExtraOperation(_InterfaceImplementationTestCase):
    @nexusrpc.service
    class Interface:
        pass

    class Impl:
        @sync_operation
        async def extra_op(self, _ctx: StartOperationContext, _input: None) -> None: ...

        def unrelated_method(self) -> None: ...

    error_message = "does not match an operation method name in the service definition"


class ValidImplWithoutTypeAnnotations(_ValidTestCase):
    @nexusrpc.service
    class Interface:
        op: nexusrpc.Operation[int, str]

    class Impl:
        @sync_operation
        async def op(self, ctx, input): ...  # type: ignore[reportMissingParameterType]


class MissingOperation(_InterfaceImplementationTestCase):
    @nexusrpc.service
    class Interface:
        op: nexusrpc.Operation[None, None]

    class Impl:
        pass

    error_message = "does not implement an operation with method name 'op'"


class MissingInputAnnotation(_ValidTestCase):
    @nexusrpc.service
    class Interface:
        op: nexusrpc.Operation[None, None]

    class Impl:
        @sync_operation
        async def op(self, ctx: StartOperationContext, input) -> None: ...  # type: ignore[reportMissingParameterType]


class MissingContextAnnotation(_ValidTestCase):
    @nexusrpc.service
    class Interface:
        op: nexusrpc.Operation[None, None]

    class Impl:
        @sync_operation
        async def op(self, ctx, input: None) -> None: ...  # type: ignore[reportMissingParameterType]


class WrongOutputType(_InvalidOutputTestCase):
    @nexusrpc.service
    class Interface:
        op: nexusrpc.Operation[None, int]

    class Impl:
        @sync_operation
        async def op(self, _ctx: StartOperationContext, _input: None) -> str: ...


class WrongOutputTypeWithNone(_InvalidOutputTestCase):
    @nexusrpc.service
    class Interface:
        op: nexusrpc.Operation[str, None]

    class Impl:
        @sync_operation
        async def op(self, _ctx: StartOperationContext, _input: str) -> str: ...


class ValidImplWithNone(_ValidTestCase):
    @nexusrpc.service
    class Interface:
        op: nexusrpc.Operation[str, None]

    class Impl:
        @sync_operation
        async def op(self, _ctx: StartOperationContext, _input: str) -> None: ...


class X:
    pass


class SuperClass:
    pass


class Subclass(SuperClass):
    pass


class OutputCovarianceImplOutputCannotBeSubclass(_InvalidOutputTestCase):
    @nexusrpc.service
    class Interface:
        op: nexusrpc.Operation[X, SuperClass]

    class Impl:
        @sync_operation
        async def op(self, _ctx: StartOperationContext, _input: X) -> Subclass: ...


class OutputCovarianceImplOutputCannotBeStrictSuperclass(_InvalidOutputTestCase):
    @nexusrpc.service
    class Interface:
        op: nexusrpc.Operation[X, Subclass]

    class Impl:
        @sync_operation
        async def op(self, _ctx: StartOperationContext, _input: X) -> SuperClass: ...


class InputContravarianceImplInputCannotBeSuperclass(_InvalidInputTestCase):
    @nexusrpc.service
    class Interface:
        op: nexusrpc.Operation[Subclass, X]

    class Impl:
        @sync_operation
        async def op(self, _ctx: StartOperationContext, _input: SuperClass) -> X: ...


class InputContravarianceImplInputCannotBeSubclass(_InvalidInputTestCase):
    @nexusrpc.service
    class Interface:
        op: nexusrpc.Operation[SuperClass, X]

    class Impl:
        @sync_operation
        async def op(self, _ctx: StartOperationContext, _input: Subclass) -> X: ...


class ValidImplWithGenericTypes(_ValidTestCase):
    """Validates that generic types work with equality comparison."""

    @nexusrpc.service
    class Interface:
        op: nexusrpc.Operation[list[int], dict[str, bool]]

    class Impl:
        @sync_operation
        async def op(
            self, _ctx: StartOperationContext, _input: list[int]
        ) -> dict[str, bool]: ...


class InvalidImplWithWrongGenericInputType(_InvalidInputTestCase):
    """Validates that mismatched generic input types are caught."""

    @nexusrpc.service
    class Interface:
        op: nexusrpc.Operation[list[int], str]

    class Impl:
        @sync_operation
        async def op(self, _ctx: StartOperationContext, _input: list[str]) -> str: ...


class InvalidImplWithWrongGenericOutputType(_InvalidOutputTestCase):
    """Validates that mismatched generic output types are caught."""

    @nexusrpc.service
    class Interface:
        op: nexusrpc.Operation[str, dict[str, int]]

    class Impl:
        @sync_operation
        async def op(
            self, _ctx: StartOperationContext, _input: str
        ) -> dict[str, str]: ...


class ValidImplWithAnyTypes(_ValidTestCase):
    """Validates that Any types require exact match (Any == Any)."""

    @nexusrpc.service
    class Interface:
        op: nexusrpc.Operation[Any, Any]

    class Impl:
        @sync_operation
        async def op(self, _ctx: StartOperationContext, _input: Any) -> Any: ...


class InvalidImplWithAnyInputButSpecificImpl(_InvalidInputTestCase):
    """Validates that Any in interface does not act as wildcard for input types."""

    @nexusrpc.service
    class Interface:
        op: nexusrpc.Operation[Any, str]

    class Impl:
        @sync_operation
        async def op(self, _ctx: StartOperationContext, _input: str) -> str: ...


class InvalidImplWithAnyOutputButSpecificImpl(_InvalidOutputTestCase):
    """Validates that Any in interface does not act as wildcard for output types."""

    @nexusrpc.service
    class Interface:
        op: nexusrpc.Operation[str, Any]

    class Impl:
        @sync_operation
        async def op(self, _ctx: StartOperationContext, _input: str) -> str: ...


@pytest.mark.parametrize(
    "test_case",
    [
        ValidImpl,
        ValidImplWithEmptyInterfaceAndExtraOperation,
        ValidImplWithoutTypeAnnotations,
        MissingOperation,
        MissingInputAnnotation,
        MissingContextAnnotation,
        WrongOutputType,
        WrongOutputTypeWithNone,
        ValidImplWithNone,
        OutputCovarianceImplOutputCannotBeSubclass,
        OutputCovarianceImplOutputCannotBeStrictSuperclass,
        InputContravarianceImplInputCannotBeSuperclass,
        InputContravarianceImplInputCannotBeSubclass,
        ValidImplWithGenericTypes,
        InvalidImplWithWrongGenericInputType,
        InvalidImplWithWrongGenericOutputType,
        ValidImplWithAnyTypes,
        InvalidImplWithAnyInputButSpecificImpl,
        InvalidImplWithAnyOutputButSpecificImpl,
    ],
)
def test_service_decorator_enforces_interface_implementation(
    test_case: type[_InterfaceImplementationTestCase],
):
    if test_case.error_message:
        with pytest.raises(Exception) as ei:
            service_handler(service=test_case.Interface)(test_case.Impl)
        err = ei.value
        assert test_case.error_message in str(err)
    else:
        service_handler(service=test_case.Interface)(test_case.Impl)


# TODO(preview): duplicate test?
def test_service_does_not_implement_operation_name():
    @nexusrpc.service
    class Contract:
        operation_a: nexusrpc.Operation[None, None]

    class Service:
        @sync_operation
        async def operation_b(
            self, _ctx: StartOperationContext, _input: None
        ) -> None: ...

    with pytest.raises(
        TypeError,
        match="does not match an operation method name in the service definition",
    ):
        _ = service_handler(service=Contract)(Service)



from dataclasses import dataclass
from typing import Any, Callable, cast

import pytest

import nexusrpc
from nexusrpc import LazyValue
from nexusrpc._util import get_operation, get_service_definition
from nexusrpc.handler import (
    Handler,
    StartOperationContext,
    service_handler,
    sync_operation,
)
from nexusrpc.handler._common import StartOperationResultSync
from tests.helpers import DummySerializer, TestOperationTaskCancellation


@dataclass()
class _TestCase:
    UserService: type[Any]
    # (service_name, op_name)
    supported_request: tuple[str, str]

    class UserServiceHandler:
        op: Callable[..., Any] = lambda: None

        async def _op_impl(self, ctx: StartOperationContext, _input: None) -> bool:
            assert (service_defn := get_service_definition(self.__class__))
            assert ctx.service == service_defn.name
            op_handler_op_defn = get_operation(self.op)
            assert op_handler_op_defn
            assert service_defn.operation_definitions.get(ctx.operation)
            return True


class NoOverrides(_TestCase):
    @nexusrpc.service
    class UserService:
        op: nexusrpc.Operation[None, bool]

    @service_handler(service=UserService)
    class UserServiceHandler(_TestCase.UserServiceHandler):
        @sync_operation
        async def op(self, ctx: StartOperationContext, input: None) -> bool:
            return await self._op_impl(ctx, input)

    supported_request = ("UserService", "op")


class OverrideServiceName(_TestCase):
    @nexusrpc.service(name="UserService-renamed")
    class UserService:
        op: nexusrpc.Operation[None, bool]

    @service_handler(service=UserService)
    class UserServiceHandler(_TestCase.UserServiceHandler):
        @sync_operation
        async def op(self, ctx: StartOperationContext, input: None) -> bool:
            return await self._op_impl(ctx, input)

    supported_request = ("UserService-renamed", "op")


class OverrideOperationName(_TestCase):
    @nexusrpc.service
    class UserService:
        op: nexusrpc.Operation[None, bool] = nexusrpc.Operation(name="op-renamed")

    @service_handler(service=UserService)
    class UserServiceHandler(_TestCase.UserServiceHandler):
        @sync_operation
        async def op(self, ctx: StartOperationContext, input: None) -> bool:
            return await self._op_impl(ctx, input)

    supported_request = ("UserService", "op-renamed")


class OverrideServiceAndOperationName(_TestCase):
    @nexusrpc.service(name="UserService-renamed")
    class UserService:
        op: nexusrpc.Operation[None, bool] = nexusrpc.Operation(name="op-renamed")

    @service_handler(service=UserService)
    class UserServiceHandler(_TestCase.UserServiceHandler):
        @sync_operation
        async def op(self, ctx: StartOperationContext, input: None) -> bool:
            return await self._op_impl(ctx, input)

    supported_request = ("UserService-renamed", "op-renamed")


@pytest.mark.parametrize(
    "test_case",
    [
        NoOverrides,
        OverrideServiceName,
        OverrideOperationName,
        OverrideServiceAndOperationName,
    ],
)
@pytest.mark.asyncio
async def test_request_routing_with_service_definition(
    test_case: _TestCase,
):
    request_service, request_op = test_case.supported_request
    ctx = StartOperationContext(
        service=request_service,
        operation=request_op,
        headers={},
        request_id="request-id",
        task_cancellation=TestOperationTaskCancellation(),
    )
    handler = Handler(user_service_handlers=[test_case.UserServiceHandler()])
    result = await handler.start_operation(
        ctx, LazyValue(serializer=DummySerializer(None), headers={})
    )
    assert cast(StartOperationResultSync[bool], result).value is True


class NoServiceDefinitionNoOverrides(_TestCase):
    @service_handler
    class UserServiceHandler(_TestCase.UserServiceHandler):
        @sync_operation
        async def op(self, ctx: StartOperationContext, input: None) -> bool:
            return await self._op_impl(ctx, input)

    supported_request = ("UserServiceHandler", "op")


class NoServiceDefinitionOverrideServiceName(_TestCase):
    @service_handler(name="UserServiceHandler-renamed")
    class UserServiceHandler(_TestCase.UserServiceHandler):
        @sync_operation
        async def op(self, ctx: StartOperationContext, input: None) -> bool:
            return await self._op_impl(ctx, input)

    supported_request = ("UserServiceHandler-renamed", "op")


class NoServiceDefinitionOverrideOperationName(_TestCase):
    @service_handler
    class UserServiceHandler(_TestCase.UserServiceHandler):
        @sync_operation(name="op-renamed")
        async def op(self, ctx: StartOperationContext, input: None) -> bool:
            return await self._op_impl(ctx, input)

    supported_request = ("UserServiceHandler", "op-renamed")


class NoServiceDefinitionOverrideServiceAndOperationName(_TestCase):
    @service_handler(name="UserServiceHandler-renamed")
    class UserServiceHandler(_TestCase.UserServiceHandler):
        @sync_operation(name="op-renamed")
        async def op(self, ctx: StartOperationContext, input: None) -> bool:
            return await self._op_impl(ctx, input)

    supported_request = ("UserServiceHandler-renamed", "op-renamed")


@pytest.mark.parametrize(
    "test_case",
    [
        NoServiceDefinitionNoOverrides,
        NoServiceDefinitionOverrideServiceName,
        NoServiceDefinitionOverrideOperationName,
        NoServiceDefinitionOverrideServiceAndOperationName,
    ],
)
@pytest.mark.asyncio
async def test_request_routing_without_service_definition(
    test_case: _TestCase,
):
    request_service, request_op = test_case.supported_request
    ctx = StartOperationContext(
        service=request_service,
        operation=request_op,
        headers={},
        request_id="request-id",
        task_cancellation=TestOperationTaskCancellation(),
    )
    handler = Handler(user_service_handlers=[test_case.UserServiceHandler()])
    result = await handler.start_operation(
        ctx, LazyValue(serializer=DummySerializer(None), headers={})
    )
    assert cast(StartOperationResultSync[bool], result).value is True



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from logging import getLogger
from typing import Any, Mapping, TypeVar

from typing_extensions import Never

logger = getLogger(__name__)

InputT = TypeVar("InputT", contravariant=True)
"""Operation input type"""

OutputT = TypeVar("OutputT", covariant=True)
"""Operation output type"""

ServiceHandlerT = TypeVar("ServiceHandlerT")
"""A user's service handler class, typically decorated with @service_handler"""

ServiceT = TypeVar("ServiceT")
"""A user's service definition class, typically decorated with @service"""


@dataclass
class Failure:
    """
    A Nexus Failure represents protocol-level failures.

    See https://github.com/nexus-rpc/api/blob/main/SPEC.md#failure
    """

    message: str
    stack_trace: str | None = None
    metadata: Mapping[str, str] | None = None
    details: Mapping[str, Any] | None = None
    cause: Failure | None = None


class HandlerError(Exception):
    """
    A Nexus handler error.

    This exception is used to represent errors that occur during the handling of a
    Nexus operation that should be reported to the caller as a handler error.

    Example:
        .. code-block:: python

            import nexusrpc

            # Raise a bad request error
            raise nexusrpc.HandlerError(
                "Invalid input provided",
                type=nexusrpc.HandlerErrorType.BAD_REQUEST
            )

            # Raise a retryable internal error
            raise nexusrpc.HandlerError(
                "Database unavailable",
                type=nexusrpc.HandlerErrorType.INTERNAL,
                retryable_override=True
            )
    """

    def __init__(
        self,
        message: str,
        *,
        type: HandlerErrorType | str,
        retryable_override: bool | None = None,
        stack_trace: str | None = None,
        original_failure: Failure | None = None,
    ):
        """
        Initialize a new HandlerError.

        :param message: A descriptive message for the error.

        :param type: The :py:class:`HandlerErrorType` of the error, or a
                     string representation of the error type. If a string is
                     provided and doesn't match a known error type, it will
                     be treated as UNKNOWN and a warning will be logged.

        :param retryable_override: Optionally set whether the error should be
                                   retried. By default, the error type is used
                                   to determine this.

        :param stack_trace: An optional stack trace string.

        :param original_failure: Set if this error is constructed from a failure object.
        """
        # Handle string error types (must be done before super().__init__ to build details)
        if isinstance(type, str):
            raw_error_type = type
            try:
                type = HandlerErrorType[type]
            except KeyError:
                logger.warning(f"Unknown Nexus HandlerErrorType: {type}")
                type = HandlerErrorType.UNKNOWN
        else:
            raw_error_type = type.value

        self.message = message
        self.type = type
        self.raw_error_type = raw_error_type
        self.retryable_override = retryable_override
        self.stack_trace = stack_trace
        self.original_failure = original_failure
        super().__init__(message)

    @property
    def retryable(self) -> bool:
        """
        Whether this error should be retried.

        If :py:attr:`retryable_override` is None, then the default behavior for the
        error type is used. See
        https://github.com/nexus-rpc/api/blob/main/SPEC.md#predefined-handler-errors
        """
        if self.retryable_override is not None:
            return self.retryable_override

        match self.type:
            case (
                HandlerErrorType.BAD_REQUEST
                | HandlerErrorType.UNAUTHENTICATED
                | HandlerErrorType.UNAUTHORIZED
                | HandlerErrorType.NOT_FOUND
                | HandlerErrorType.CONFLICT
                | HandlerErrorType.NOT_IMPLEMENTED
            ):
                return False
            case (
                HandlerErrorType.RESOURCE_EXHAUSTED
                | HandlerErrorType.REQUEST_TIMEOUT
                | HandlerErrorType.INTERNAL
                | HandlerErrorType.UNAVAILABLE
                | HandlerErrorType.UPSTREAM_TIMEOUT
                | HandlerErrorType.UNKNOWN
            ):
                return True

            # Type checking enforces exhaustive matching but
            # the default case is included to provide a runtime default.
            # If a case is missing from above, the assignment to Never
            # will cause a type checking error.
            case _ as unreachable:  # pyright: ignore[reportUnnecessaryComparison]
                _: Never = unreachable  # pyright: ignore[reportUnreachable]
                return True  # pyright: ignore[reportUnreachable]


class HandlerErrorType(Enum):
    """Nexus handler error types.

    See https://github.com/nexus-rpc/api/blob/main/SPEC.md#predefined-handler-errors
    """

    UNKNOWN = "UNKNOWN"
    """
    The error type is unknown. Subsequent requests by the client are permissible.
    """

    BAD_REQUEST = "BAD_REQUEST"
    """
    The handler cannot or will not process the request due to an apparent client error.

    Clients should not retry this request unless advised otherwise.
    """

    UNAUTHENTICATED = "UNAUTHENTICATED"
    """
    The client did not supply valid authentication credentials for this request.

    Clients should not retry this request unless advised otherwise.
    """

    UNAUTHORIZED = "UNAUTHORIZED"
    """
    The caller does not have permission to execute the specified operation.

    Clients should not retry this request unless advised otherwise.
    """

    NOT_FOUND = "NOT_FOUND"
    """
    The requested resource could not be found but may be available in the future.

    Subsequent requests by the client are permissible but not advised.
    """

    REQUEST_TIMEOUT = "REQUEST_TIMEOUT"
    """
    Returned by the server when it has given up handling a request. This may occur by enforcing a client
    provided `Request-Timeout` or for any arbitrary reason such as enforcing some configurable limit.

    Subsequent requests by the client are permissible.
    """

    CONFLICT = "CONFLICT"
    """
    The request could not be made due to a conflict. This may happen when trying to create an operation that
    has already been started.

    Clients should not retry this request unless advised otherwise.
    """

    RESOURCE_EXHAUSTED = "RESOURCE_EXHAUSTED"
    """
    Some resource has been exhausted, perhaps a per-user quota, or perhaps the entire file system is out of space.

    Subsequent requests by the client are permissible.
    """

    INTERNAL = "INTERNAL"
    """
    An internal error occurred.

    Subsequent requests by the client are permissible.
    """

    NOT_IMPLEMENTED = "NOT_IMPLEMENTED"
    """
    The handler either does not recognize the request method, or it lacks the ability to fulfill the request.

    Clients should not retry this request unless advised otherwise.
    """

    UNAVAILABLE = "UNAVAILABLE"
    """
    The service is currently unavailable.

    Subsequent requests by the client are permissible.
    """

    UPSTREAM_TIMEOUT = "UPSTREAM_TIMEOUT"
    """
    Used by gateways to report that a request to an upstream handler has timed out.

    Subsequent requests by the client are permissible.
    """


class OperationError(Exception):
    """
    An error that represents "failed" and "canceled" operation results.

    Example:
        .. code-block:: python

            import nexusrpc

            # Indicate operation failed
            raise nexusrpc.OperationError(
                "Processing failed due to invalid data",
                state=nexusrpc.OperationErrorState.FAILED
            )

            # Indicate operation was canceled
            raise nexusrpc.OperationError(
                "Operation was canceled by user request",
                state=nexusrpc.OperationErrorState.CANCELED
            )
    """

    def __init__(
        self,
        message: str,
        *,
        state: OperationErrorState,
        stack_trace: str | None = None,
        original_failure: Failure | None = None,
    ):
        """
        Initialize a new OperationError.

        :param message: A descriptive message for the error.

        :param state: The state of the operation (:py:class:`OperationErrorState`).

        :param stack_trace: An optional stack trace string.

        :param original_failure: Set if this error is constructed from a failure object.

        """

        self.message = message
        self.state = state
        self.stack_trace = stack_trace
        self.original_failure = original_failure

        self.state = state
        super().__init__(message)


class OperationErrorState(Enum):
    """
    The state of an operation as described by an :py:class:`OperationError`.
    """

    FAILED = "failed"
    """
    The operation failed.
    """

    CANCELED = "canceled"
    """
    The operation was canceled.
    """


@dataclass(frozen=True)
class Link:
    """
    A Link contains a URL and a type that can be used to decode the URL.

    The URL may contain arbitrary data (percent-encoded). It can be used to pass
    information about the caller to the handler, or vice versa.
    """

    url: str
    """
    Link URL.

    Must be percent-encoded.
    """

    type: str
    """
    A data type for decoding the URL.

    Valid chars: alphanumeric, '_', '.', '/'
    """



from dataclasses import dataclass
from typing import Optional

import pytest

import nexusrpc
from nexusrpc._util import get_service_definition
from nexusrpc.handler import service_handler


@nexusrpc.service
class ServiceInterface:
    pass


@nexusrpc.service(name="Service-With-Name-Override")
class ServiceInterfaceWithNameOverride:
    pass


@dataclass()
class _NameOverrideTestCase:
    ServiceImpl: type
    expected_name: str
    expected_error: Optional[type[Exception]] = None


class NotCalled(_NameOverrideTestCase):
    @service_handler
    class ServiceImpl:
        pass

    expected_name = "ServiceImpl"


class CalledWithoutArgs(_NameOverrideTestCase):
    @service_handler()
    class ServiceImpl:
        pass

    expected_name = "ServiceImpl"


class CalledWithNameArg(_NameOverrideTestCase):
    @service_handler(name="my-service-impl-🌈")
    class ServiceImpl:
        pass

    expected_name = "my-service-impl-🌈"


class CalledWithInterface(_NameOverrideTestCase):
    @service_handler(service=ServiceInterface)
    class ServiceImpl:
        pass

    expected_name = "ServiceInterface"


class CalledWithInterfaceWithNameOverride(_NameOverrideTestCase):
    @service_handler(service=ServiceInterfaceWithNameOverride)
    class ServiceImpl:
        pass

    expected_name = "Service-With-Name-Override"


@pytest.mark.parametrize(
    "test_case",
    [
        NotCalled,
        CalledWithoutArgs,
        CalledWithNameArg,
        CalledWithInterface,
        CalledWithInterfaceWithNameOverride,
    ],
)
def test_service_decorator_name_overrides(test_case: type[_NameOverrideTestCase]):
    service = get_service_definition(test_case.ServiceImpl)
    assert isinstance(service, nexusrpc.ServiceDefinition)
    assert service.name == test_case.expected_name


def test_name_must_not_be_empty():
    with pytest.raises(ValueError):
        _ = service_handler(name="")(object)


def test_name_and_interface_are_mutually_exclusive():
    with pytest.raises(ValueError):
        # Type error due to deliberately violating overload
        service_handler(name="my-service-impl-🌈", service=ServiceInterface)  # type: ignore



from dataclasses import dataclass
from typing import Any

import pytest

from nexusrpc.handler import (
    OperationHandler,
    service_handler,
)
from nexusrpc.handler._decorators import operation_handler


@dataclass()
class _TestCase:
    UserServiceHandler: type[Any]
    expected_error_message: str


class DuplicateOperationName(_TestCase):
    class UserServiceHandler:
        @operation_handler(name="a")
        def op_1(self) -> OperationHandler[int, int]: ...

        @operation_handler(name="a")
        def op_2(self) -> OperationHandler[str, int]: ...

    expected_error_message = (
        "Operation 'a' in service 'UserServiceHandler' is defined multiple times."
    )


@pytest.mark.parametrize(
    "test_case",
    [
        DuplicateOperationName,
    ],
)
def test_service_handler_decorator(test_case: _TestCase):
    with pytest.raises(RuntimeError, match=test_case.expected_error_message):
        service_handler(test_case.UserServiceHandler)



from unittest import mock

import pytest

from nexusrpc._util import get_operation_factory, is_async_callable
from nexusrpc.handler import (
    StartOperationContext,
    StartOperationResultSync,
    service_handler,
    sync_operation,
)


@service_handler
class MyServiceHandler:
    def __init__(self):
        self.mutable_container = []

    @sync_operation
    def my_def_op(self, _ctx: StartOperationContext, input: int) -> int:
        """
        This is the docstring for the `my_def_op` sync operation.
        """
        self.mutable_container.append(input)
        return input + 1

    @sync_operation(name="foo")
    async def my_async_def_op(self, _ctx: StartOperationContext, input: int) -> int:
        """
        This is the docstring for the `my_async_def_op` sync operation.
        """
        self.mutable_container.append(input)
        return input + 2


def test_def_sync_handler():
    user_instance = MyServiceHandler()
    op_handler_factory = get_operation_factory(user_instance.my_def_op)
    assert op_handler_factory
    op_handler = op_handler_factory(user_instance)
    assert not is_async_callable(op_handler.start)
    assert (
        str(op_handler.start.__doc__).strip()
        == "This is the docstring for the `my_def_op` sync operation."
    )
    assert not user_instance.mutable_container
    ctx = mock.Mock(spec=StartOperationContext)
    result = op_handler.start(ctx, 1)
    assert isinstance(result, StartOperationResultSync)
    assert result.value == 2
    assert user_instance.mutable_container == [1]


@pytest.mark.asyncio
async def test_async_def_sync_handler():
    user_instance = MyServiceHandler()
    op_handler_factory = get_operation_factory(user_instance.my_async_def_op)
    assert op_handler_factory
    op_handler = op_handler_factory(user_instance)
    assert is_async_callable(op_handler.start)
    assert (
        str(op_handler.start.__doc__).strip()
        == "This is the docstring for the `my_async_def_op` sync operation."
    )
    assert not user_instance.mutable_container
    ctx = mock.Mock(spec=StartOperationContext)
    result = await op_handler.start(ctx, 1)
    assert isinstance(result, StartOperationResultSync)
    assert result.value == 3
    assert user_instance.mutable_container == [1]



from nexusrpc._common import (
    HandlerError,
    HandlerErrorType,
)


def test_handler_error_retryable_behavior():
    """Test retryable behavior based on error type and override."""
    # Retryable error type (RESOURCE_EXHAUSTED)
    retryable_type = HandlerErrorType.RESOURCE_EXHAUSTED
    err = HandlerError("test", type=retryable_type)
    assert err.retryable
    assert err.type == retryable_type
    assert err.raw_error_type == retryable_type.value

    err = HandlerError("test", type=retryable_type, retryable_override=False)
    assert not err.retryable

    # Non-retryable error type (BAD_REQUEST)
    non_retryable_type = HandlerErrorType.BAD_REQUEST
    err = HandlerError("test", type=non_retryable_type)
    assert not err.retryable
    assert err.type == non_retryable_type
    assert err.raw_error_type == non_retryable_type.value

    err = HandlerError("test", type=non_retryable_type, retryable_override=True)
    assert err.retryable


def test_handler_error_unknown_error_type():
    """Test handling of unknown error type strings."""
    err = HandlerError("test", type="SOME_UNKNOWN_TYPE")
    assert err.retryable
    assert err.type == HandlerErrorType.UNKNOWN
    assert err.raw_error_type == "SOME_UNKNOWN_TYPE"

    err = HandlerError("test", type="SOME_UNKNOWN_TYPE", retryable_override=False)
    assert not err.retryable



import warnings
from collections.abc import Awaitable
from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    Union,
    get_args,
    get_origin,
)

import pytest

from nexusrpc.handler import StartOperationContext
from nexusrpc.handler._util import get_start_method_input_and_output_type_annotations


class Input:
    pass


class Output:
    pass


@dataclass()
class _TestCase:
    start: Callable[..., Any]
    expected_types: tuple[Any, Any]


class SyncMethod(_TestCase):
    async def start(self, ctx: StartOperationContext, i: Input) -> Output: ...

    expected_types = (Input, Output)


class AsyncMethod(_TestCase):
    async def start(self, ctx: StartOperationContext, i: Input) -> Output: ...

    expected_types = (Input, Output)


class UnionMethod(_TestCase):
    async def start(
        self, ctx: StartOperationContext, i: Input
    ) -> Union[Output, Awaitable[Output]]: ...

    expected_types = (Input, Union[Output, Awaitable[Output]])


class MissingInputAnnotationInUnionMethod(_TestCase):
    async def start(
        self,
        ctx: StartOperationContext,
        i,  # type: ignore[reportMissingParameterType]
    ) -> Union[Output, Awaitable[Output]]: ...

    expected_types = (None, Union[Output, Awaitable[Output]])


class TooFewParams(_TestCase):
    async def start(self, i: Input) -> Output: ...

    expected_types = (None, Output)


class TooManyParams(_TestCase):
    async def start(
        self, ctx: StartOperationContext, i: Input, extra: int
    ) -> Output: ...

    expected_types = (None, Output)


class WrongOptionsType(_TestCase):
    async def start(self, ctx: int, i: Input) -> Output: ...

    expected_types = (None, Output)


class NoReturnHint(_TestCase):
    async def start(self, ctx: StartOperationContext, i: Input): ...

    expected_types = (Input, None)


class NoInputAnnotation(_TestCase):
    async def start(self, ctx: StartOperationContext, i) -> Output: ...  # type: ignore[reportMissingParameterType]

    expected_types = (None, Output)


class NoOptionsAnnotation(_TestCase):
    async def start(self, ctx, i: Input) -> Output: ...  # type: ignore[reportMissingParameterType]

    expected_types = (None, Output)


class AllAnnotationsMissing(_TestCase):
    async def start(self, ctx: StartOperationContext, i): ...  # type: ignore[reportMissingParameterType]

    expected_types = (None, None)


class ExplicitNoneTypes(_TestCase):
    async def start(self, ctx: StartOperationContext, i: None) -> None: ...

    expected_types = (type(None), type(None))


@pytest.mark.parametrize(
    "test_case",
    [
        SyncMethod,
        AsyncMethod,
        UnionMethod,
        TooFewParams,
        TooManyParams,
        WrongOptionsType,
        NoReturnHint,
        NoInputAnnotation,
        NoOptionsAnnotation,
        MissingInputAnnotationInUnionMethod,
        AllAnnotationsMissing,
        ExplicitNoneTypes,
    ],
)
def test_get_input_and_output_types(test_case: type[_TestCase]):
    with warnings.catch_warnings(record=True):
        warnings.simplefilter("always")
        input_type, output_type = get_start_method_input_and_output_type_annotations(
            test_case.start
        )
        expected_input_type, expected_output_type = test_case.expected_types
        assert input_type is expected_input_type

        expected_origin = get_origin(expected_output_type)
        if expected_origin:  # Awaitable and Union cases
            assert get_origin(output_type) is expected_origin
            assert get_args(output_type) == get_args(expected_output_type)
        else:
            assert output_type is expected_output_type



"""
Components for implementing Nexus handlers.

Server/worker authors will use this module to create the top-level Nexus handlers
responsible for dispatching requests to Nexus operations.

Nexus service/operation authors will use this module to implement operation handler
methods within service handler classes.
"""

from __future__ import annotations

from ._common import (
    CancelOperationContext,
    OperationContext,
    OperationTaskCancellation,
    StartOperationContext,
    StartOperationResultAsync,
    StartOperationResultSync,
)
from ._core import Handler, OperationHandlerMiddleware
from ._decorators import operation_handler, service_handler, sync_operation
from ._operation_handler import MiddlewareSafeOperationHandler, OperationHandler

__all__ = [
    "MiddlewareSafeOperationHandler",
    "CancelOperationContext",
    "Handler",
    "OperationContext",
    "OperationHandler",
    "OperationTaskCancellation",
    "OperationHandlerMiddleware",
    "operation_handler",
    "service_handler",
    "StartOperationContext",
    "StartOperationResultAsync",
    "StartOperationResultSync",
    "sync_operation",
    "operation_handler",
]



"""
A Nexus service definition is a class with class attributes of type Operation. It must
be be decorated with @nexusrpc.service. The decorator validates the Operation
attributes.
"""

from __future__ import annotations

import dataclasses
import typing
from collections.abc import Mapping
from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    Generic,
    Optional,
    Union,
    overload,
)

from typing_extensions import dataclass_transform

from nexusrpc._common import InputT, OutputT, ServiceT
from nexusrpc._util import (
    get_annotations,
    get_service_definition,
    set_service_definition,
)


@dataclass
class Operation(Generic[InputT, OutputT]):
    """Defines a Nexus operation in a Nexus service definition.

    This class is for definition of operation name and input/output types only; to
    implement an operation, see `:py:meth:nexusrpc.handler.operation_handler`.

    Example:

    .. code-block:: python

        @nexusrpc.service
        class MyNexusService:
            my_operation: nexusrpc.Operation[MyInput, MyOutput]
    """

    name: str
    method_name: Optional[str] = dataclasses.field(default=None, init=False)
    input_type: Optional[type[InputT]] = dataclasses.field(default=None)
    output_type: Optional[type[OutputT]] = dataclasses.field(default=None)


@dataclass
class OperationDefinition(Generic[InputT, OutputT]):
    """
    Internal representation of a user's :py:class:`Operation` definition.
    """

    name: str
    method_name: str
    input_type: type[InputT]
    output_type: type[OutputT]

    @classmethod
    def from_operation(
        cls, operation: Operation[InputT, OutputT]
    ) -> OperationDefinition[InputT, OutputT]:
        if not operation.name:
            raise ValueError(
                f"Operation has no name (method_name is '{operation.method_name}')"
            )
        if not operation.method_name:
            raise ValueError(f"Operation '{operation.name}' has no method name")
        if not operation.input_type:
            raise ValueError(f"Operation '{operation.name}' has no input type")
        if not operation.output_type:
            raise ValueError(f"Operation '{operation.name}' has no output type")
        return cls(
            name=operation.name,
            method_name=operation.method_name,
            input_type=operation.input_type,
            output_type=operation.output_type,
        )


@overload
@dataclass_transform(field_specifiers=(Operation,))
def service(cls: type[ServiceT]) -> type[ServiceT]: ...


@overload
@dataclass_transform(field_specifiers=(Operation,))
def service(
    *, name: Optional[str] = None
) -> Callable[[type[ServiceT]], type[ServiceT]]: ...


def service(
    cls: Optional[type[ServiceT]] = None,
    *,
    name: Optional[str] = None,
) -> Union[
    type[ServiceT],
    Callable[[type[ServiceT]], type[ServiceT]],
]:
    """
    Decorator marking a class as a Nexus service definition.

    The decorator validates the operation definitions in the service definition: that they
    have the correct type, and that there are no duplicate operation names. The decorator
    also creates instances of the Operation class for each operation definition.

    Example:
        .. code-block:: python

            @nexusrpc.service
            class MyNexusService:
                my_op: nexusrpc.Operation[MyInput, MyOutput]
                another_op: nexusrpc.Operation[str, dict]

            @nexusrpc.service(name="custom-service-name")
            class AnotherService:
                process: nexusrpc.Operation[ProcessInput, ProcessOutput]
    """

    def decorator(cls: type[ServiceT]) -> type[ServiceT]:
        if name is not None and not name:
            raise ValueError("Service name must not be empty.")
        defn = ServiceDefinition.from_class(cls, name or cls.__name__)
        set_service_definition(cls, defn)

        # In order for callers to refer to operation definitions at run-time, a decorated user
        # service class must itself have a class attribute for every operation, even if
        # declared only via a type annotation, and whether inherited from a parent class
        # or not.
        #
        # TODO(preview): it is sufficient to do this setattr only for the subset of
        # operations that were declared on *this* class. Currently however we are
        # setting all inherited operations.
        for op_name, op_defn in defn.operation_definitions.items():
            if not hasattr(cls, op_name):
                op = Operation(
                    name=op_defn.name,
                    input_type=op_defn.input_type,
                    output_type=op_defn.output_type,
                )
                op.method_name = op_defn.method_name
                setattr(cls, op_name, op)

        return cls

    if cls is None:
        return decorator
    else:
        return decorator(cls)


@dataclass(frozen=True)
class ServiceDefinition:
    """
    Internal representation of a user's service definition class.

    A named collection of named operation definitions.

    A service definition class is a class decorated with :py:func:`nexusrpc.service`
    containing class attribute type annotations of type :py:class:`nexusrpc.Operation`.
    """

    name: str
    operation_definitions: Mapping[str, OperationDefinition[Any, Any]]

    def __post_init__(self):
        if errors := self._validation_errors():
            raise ValueError(
                f"Service definition {self.name} has validation errors: {', '.join(errors)}"
            )

    @staticmethod
    def from_class(user_class: type[ServiceT], name: str) -> ServiceDefinition:
        """Create a ServiceDefinition from a user service definition class.

        The set of service definition operations returned is the union of operations
        defined directly on this class with those inherited from ancestral service
        definitions (i.e. ancestral classes that were decorated with @nexusrpc.service).
        If multiple service definitions define an operation with the same name, then the
        usual mro() precedence rules apply.
        """
        operation_definitions = ServiceDefinition._collect_operations(user_class)

        # Obtain the set of operations to be inherited from ancestral service
        # definitions. Operations are only inherited from classes that are also
        # decorated with @nexusrpc.service. We do not permit any "overriding" by child
        # classes; both the following must be true:
        # 1. No inherited operation has the same name as that of an operation defined
        #    here. If this were violated, it would indicate two service definitions
        #    exposing potentially different operation definitions behind the same
        #    operation name.
        # 2. No inherited operation has the same method name as that of an operation
        #    defined here. If this were violated, there would be ambiguity in which
        #    operation handler is dispatched to.
        parent_service_defn = next(
            (
                defn
                for defn in (
                    get_service_definition(cls) for cls in user_class.mro()[1:]
                )
                if defn
            ),
            None,
        )
        if parent_service_defn:
            method_names = {op.method_name for op in operation_definitions.values()}
            for op_defn in parent_service_defn.operation_definitions.values():
                if op_defn.method_name in method_names:
                    raise ValueError(
                        f"Operation method name '{op_defn.method_name}' in class '{user_class}' "
                        f"also occurs in a service definition inherited from a parent class: "
                        f"'{parent_service_defn.name}'. This is not allowed."
                    )
                if op_defn.name in operation_definitions:
                    raise ValueError(
                        f"Operation name '{op_defn.name}' in class '{user_class}' "
                        f"also occurs in a service definition inherited from a parent class: "
                        f"'{parent_service_defn.name}'. This is not allowed."
                    )
                operation_definitions[op_defn.name] = op_defn

        return ServiceDefinition(name=name, operation_definitions=operation_definitions)

    def _validation_errors(self) -> list[str]:
        errors = []
        if not self.name:
            errors.append("Service has no name")
        seen_method_names = set()
        for op_defn in self.operation_definitions.values():
            if op_defn.method_name in seen_method_names:
                errors.append(
                    f"Operation method name '{op_defn.method_name}' is not unique"
                )
            seen_method_names.add(op_defn.method_name)
        return errors

    @staticmethod
    def _collect_operations(
        user_class: type[ServiceT],
    ) -> dict[str, OperationDefinition[Any, Any]]:
        """Collect operation definitions from a user service definition class.

        Does not visit parent classes.
        """

        # Form the union of all class attribute names that are either an Operation
        # instance or have an Operation type annotation, or both.
        operations: dict[str, Operation[Any, Any]] = {}
        for k, v in user_class.__dict__.items():
            if isinstance(v, Operation):
                operations[k] = v
            elif typing.get_origin(v) is Operation:
                raise TypeError(
                    "Operation definitions in the service definition should look like  "
                    "my_op: nexusrpc.Operation[InputType, OutputType]. Did you accidentally "
                    "use '=' instead of ':'?"
                )

        annotations = {
            k: v
            for k, v in get_annotations(user_class, eval_str=True).items()
            if v == Operation or typing.get_origin(v) == Operation
        }
        for key in operations.keys() | annotations.keys():
            # If the name has a type annotation, then add the input and output types to
            # the operation instance, or create the instance if there was only an
            # annotation.
            if op_type := annotations.get(key):
                args = typing.get_args(op_type)
                if len(args) != 2:
                    raise TypeError(
                        f"Operation types in the service definition should look like  "
                        f"nexusrpc.Operation[InputType, OutputType], but '{key}' in "
                        f"'{user_class}' has {len(args)} type parameters."
                    )
                input_type, output_type = args
                if key not in operations:
                    # It looked like
                    # my_op: Operation[I, O]
                    op = operations[key] = Operation(
                        name=key,
                        input_type=input_type,
                        output_type=output_type,
                    )
                else:
                    op = operations[key]
                    # It looked like
                    # my_op: Operation[I, O] = Operation(...)
                    if not op.input_type:
                        op.input_type = input_type
                    elif op.input_type != input_type:
                        raise ValueError(
                            f"Operation {key} input_type ({op.input_type}) must match type parameter {input_type}"
                        )
                    if not op.output_type:
                        op.output_type = output_type
                    elif op.output_type != output_type:
                        raise ValueError(
                            f"Operation {key} output_type ({op.output_type}) must match type parameter {output_type}"
                        )
            else:
                # It looked like
                # my_op = Operation(...)
                op = operations[key]

            # Validate that if method_name was set (via direct assignment after
            # construction), it matches the attribute name
            if op.method_name is not None and op.method_name != key:
                raise ValueError(
                    f"Operation {key} method_name ({op.method_name}) must match attribute name {key}"
                )
            op.method_name = key

        op_defns = {}
        for op in operations.values():
            if op.name in op_defns:
                raise RuntimeError(
                    f"Operation '{op.name}' in service '{user_class}' is defined multiple times"
                )
            op_defns[op.name] = OperationDefinition.from_operation(op)
        return op_defns



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Mapping, Sequence
from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, Generic, Optional

from nexusrpc._common import Link, OutputT


class OperationTaskCancellation(ABC):
    """
    Indicates whether a a Nexus task has been cancelled during a sync operation or before an async operation has
    returned a token.

    Nexus worker implementations are expected to provide an implementation that enables
    cooperative cancellation for both sync and async operation handlers.

    Operation Handler implementations are expected to periodically check :py:attr:`is_cancelled` or use :py:attr:`wait_until_cancelled` / :py:attr:`wait_until_cancelled_sync` to cancel in flight work when appropriate.
    """

    @abstractmethod
    def is_cancelled(self) -> bool:
        """
        Return True if the associated task has been cancelled.
        """
        raise NotImplementedError

    @abstractmethod
    def cancellation_reason(self) -> Optional[str]:
        """
        Provide additional context for the cancellation, if available.
        """
        raise NotImplementedError

    @abstractmethod
    def wait_until_cancelled_sync(self, timeout: Optional[float] = None) -> bool:
        """
        Block until cancellation occurs or the optional timeout elapses. Nexus worker implementations may return `True` for :py:attr:`is_cancelled` before this method returns and therefore may cause a race condition if both are used in tandem.
        """
        raise NotImplementedError

    @abstractmethod
    async def wait_until_cancelled(self) -> None:
        """Await cancellation using async primitives. Nexus worker implementations may return `True` for :py:attr:`is_cancelled` before this method returns and therefore may cause a race condition if both are used in tandem."""
        raise NotImplementedError


@dataclass(frozen=True, kw_only=True)
class OperationContext(ABC):
    """
    Context for the execution of the requested operation method.

    Includes information from the request.
    """

    def __new__(cls, *args: Any, **kwargs: Any):
        if cls is OperationContext:
            raise TypeError(
                "OperationContext is an abstract class and cannot be instantiated directly"
            )
        return super().__new__(cls)

    service: str
    """
    The name of the service that the operation belongs to.
    """

    operation: str
    """
    The name of the operation.
    """

    headers: Mapping[str, str]
    """
    Optional header fields sent by the caller.
    """

    task_cancellation: OperationTaskCancellation
    """
    Task cancellation information indicating that a running task should be interrupted. This is distinct from operation cancellation.
    """

    request_deadline: Optional[datetime] = None
    """
    The deadline for the operation handler method. Note that this is the time by which the
    current _request_ should complete, not the _operation_'s deadline.
    """


@dataclass(frozen=True, kw_only=True)
class StartOperationContext(OperationContext):
    """
    Context for the start method.

    Includes information from the request.
    """

    request_id: str
    """
    Request ID that may be used by the handler to dedupe a start request.
    By default a v4 UUID should be generated by the client.
    """

    callback_url: Optional[str] = None
    """
    A callback URL is required to deliver the completion of an async operation. This URL should be
    called by a handler upon completion if the started operation is async.
    """

    callback_headers: Mapping[str, str] = field(default_factory=dict)
    """
    Optional header fields set by the caller to be attached to the callback request when an
    asynchronous operation completes.
    """

    inbound_links: Sequence[Link] = field(default_factory=list)
    """
    Links received in the request. This list is automatically populated when handling a start
    request. Handlers may use these links, for example to add information about the
    caller to a resource associated with the operation execution.
    """

    outbound_links: list[Link] = field(default_factory=list)
    """
    Links to be returned by the handler. This list is initially empty. Handlers may
    populate this list, for example with links describing resources associated with the
    operation execution that may be of use to the caller.
    """


@dataclass(frozen=True, kw_only=True)
class CancelOperationContext(OperationContext):
    """
    Context for the cancel method.

    Includes information from the request.
    """


@dataclass(frozen=True)
class StartOperationResultSync(Generic[OutputT]):
    """
    A result returned synchronously by the start method of a nexus operation handler.
    """

    value: OutputT  # type: ignore[misc]
    """
    The value returned by the operation.
    """


@dataclass(frozen=True)
class StartOperationResultAsync:
    """
    A value returned by the start method of a nexus operation handler indicating that
    the operation is responding asynchronously.
    """

    token: str
    """
    A token representing the in-progress operation that the caller can submit with
    subsequent ``fetch_info``, ``fetch_result``, or ``cancel`` requests.
    """



from __future__ import annotations

import functools
import inspect
from collections.abc import Awaitable
from typing import TYPE_CHECKING, Any, Callable, Optional

from typing_extensions import TypeGuard

import nexusrpc

if TYPE_CHECKING:
    import nexusrpc
    from nexusrpc import InputT, OutputT
    from nexusrpc._common import ServiceT
    from nexusrpc.handler._operation_handler import OperationHandler

_NEXUS_SERVICE_DEFINITION_ATTR_NAME = "__nexus_service_definition__"
_NEXUS_OPERATION_ATTR_NAME = "__nexus_operation__"
_NEXUS_OPERATION_FACTORY_ATTR_NAME = "__nexus_operation_factory__"


def get_service_definition(
    obj: Any,
) -> Optional[nexusrpc.ServiceDefinition]:
    """Return the :py:class:`nexusrpc.ServiceDefinition` for the object, or None"""
    # Do not use getattr since it would allow a non-decorated class to act as a service
    # definition if it inherits from a decorated class.
    if isinstance(obj, type):
        defn = obj.__dict__.get(_NEXUS_SERVICE_DEFINITION_ATTR_NAME)
    else:
        defn = getattr(obj, "__dict__", {}).get(_NEXUS_SERVICE_DEFINITION_ATTR_NAME)
    if defn and not isinstance(defn, nexusrpc.ServiceDefinition):
        raise ValueError(
            f"{obj.__name__} has a {_NEXUS_SERVICE_DEFINITION_ATTR_NAME} attribute "
            f"that is not a nexusrpc.ServiceDefinition."
        )
    return defn


def set_service_definition(
    cls: type[ServiceT], service_definition: nexusrpc.ServiceDefinition
) -> None:
    """Set the :py:class:`nexusrpc.ServiceDefinition` for this class."""
    setattr(cls, _NEXUS_SERVICE_DEFINITION_ATTR_NAME, service_definition)


def get_operation(
    obj: Any,
) -> Optional[nexusrpc.Operation[Any, Any]]:
    """Return the :py:class:`nexusrpc.Operation` for the object, or None

    ``obj`` should be a decorated operation start method, or a method that takes
    no arguments and returns an OperationHandler.
    """
    if factory := getattr(obj, _NEXUS_OPERATION_FACTORY_ATTR_NAME, None):
        # obj was a decorated operation start method
        obj = factory
    op = getattr(obj, _NEXUS_OPERATION_ATTR_NAME, None)
    if op and not isinstance(op, nexusrpc.Operation):
        raise ValueError(f"{op} is not a nexusrpc.Operation")
    return op


def set_operation(
    obj: Any,
    operation: nexusrpc.Operation[Any, Any],
) -> None:
    """Set the :py:class:`nexusrpc.Operation` for this object.

    ``obj`` should be an operation start method.
    """
    if not isinstance(operation, nexusrpc.Operation):  # type: ignore
        raise ValueError(f"{operation} is not a nexusrpc.Operation")  # type: ignore
    setattr(obj, _NEXUS_OPERATION_ATTR_NAME, operation)


def get_operation_factory(
    obj: Any,
) -> Optional[Callable[[Any], OperationHandler[Any, Any]]]:
    """Return the :py:class:`OperationHandler` factory function for the object."""
    if factory := getattr(obj, _NEXUS_OPERATION_FACTORY_ATTR_NAME, None):
        # obj was a decorated operation start method
        return factory
    if get_operation(obj):
        # obj was the desired factory
        return obj
    return None


def set_operation_factory(
    obj: Any,
    operation_factory: Callable[[Any], OperationHandler[InputT, OutputT]],
) -> None:
    """Set the :py:class:`OperationHandler` factory for this object.

    ``obj`` should be an operation start method.
    """
    setattr(obj, _NEXUS_OPERATION_FACTORY_ATTR_NAME, operation_factory)


# Copied from https://github.com/modelcontextprotocol/python-sdk
#
# Copyright (c) 2024 Anthropic, PBC.
#
# Modified to use TypeGuard.
#
# This file is licensed under the MIT License.
def is_async_callable(obj: Any) -> TypeGuard[Callable[..., Awaitable[Any]]]:
    """
    Return True if `obj` is an async callable.

    Supports partials of async callable class instances.
    """
    while isinstance(obj, functools.partial):
        obj = obj.func

    return inspect.iscoroutinefunction(obj) or (
        callable(obj) and inspect.iscoroutinefunction(getattr(obj, "__call__", None))
    )


def is_callable(obj: Any) -> TypeGuard[Callable[..., Any]]:
    """
    Return True if `obj` is a callable.
    """
    while isinstance(obj, functools.partial):
        obj = obj.func
    return inspect.isfunction(obj) or (callable(obj) and hasattr(obj, "__call__"))


def get_callable_name(fn: Callable[..., Any]) -> str:
    method_name = getattr(fn, "__name__", None)
    if not method_name and callable(fn) and hasattr(fn, "__call__"):
        method_name = fn.__class__.__name__
    if not method_name:
        raise TypeError(
            f"Could not determine callable name: "
            f"expected {fn} to be a function or callable instance."
        )
    return method_name


# See
# https://docs.python.org/3/howto/annotations.html#accessing-the-annotations-dict-of-an-object-in-python-3-9-and-older

try:
    from inspect import (  # type: ignore
        get_annotations as get_annotations,  # type: ignore[reportAttributeAccessIssue]
    )
except ImportError:
    import functools
    import sys
    import types

    # This is inspect.get_annotations from Python 3.13.5
    def get_annotations(obj, *, globals=None, locals=None, eval_str=False):  # type: ignore[misc]
        """Compute the annotations dict for an object.

        obj may be a callable, class, or module.
        Passing in an object of any other type raises TypeError.

        Returns a dict.  get_annotations() returns a new dict every time
        it's called; calling it twice on the same object will return two
        different but equivalent dicts.

        This function handles several details for you:

        * If eval_str is true, values of type str will
            be un-stringized using eval().  This is intended
            for use with stringized annotations
            ("from __future__ import annotations").
        * If obj doesn't have an annotations dict, returns an
            empty dict.  (Functions and methods always have an
            annotations dict; classes, modules, and other types of
            callables may not.)
        * Ignores inherited annotations on classes.  If a class
            doesn't have its own annotations dict, returns an empty dict.
        * All accesses to object members and dict values are done
            using getattr() and dict.get() for safety.
        * Always, always, always returns a freshly-created dict.

        eval_str controls whether or not values of type str are replaced
        with the result of calling eval() on those values:

        * If eval_str is true, eval() is called on values of type str.
        * If eval_str is false (the default), values of type str are unchanged.

        globals and locals are passed in to eval(); see the documentation
        for eval() for more information.  If either globals or locals is
        None, this function may replace that value with a context-specific
        default, contingent on type(obj):

        * If obj is a module, globals defaults to obj.__dict__.
        * If obj is a class, globals defaults to
            sys.modules[obj.__module__].__dict__ and locals
            defaults to the obj class namespace.
        * If obj is a callable, globals defaults to obj.__globals__,
            although if obj is a wrapped function (using
            functools.update_wrapper()) it is first unwrapped.
        """
        if isinstance(obj, type):
            # class
            obj_dict = getattr(obj, "__dict__", None)
            if obj_dict and hasattr(obj_dict, "get"):
                ann = obj_dict.get("__annotations__", None)
                if isinstance(ann, types.GetSetDescriptorType):
                    ann = None
            else:
                ann = None

            obj_globals = None
            module_name = getattr(obj, "__module__", None)
            if module_name:
                module = sys.modules.get(module_name, None)
                if module:
                    obj_globals = getattr(module, "__dict__", None)
            obj_locals = dict(vars(obj))
            unwrap = obj
        elif isinstance(obj, types.ModuleType):
            # module
            ann = getattr(obj, "__annotations__", None)
            obj_globals = getattr(obj, "__dict__")
            obj_locals = None
            unwrap = None
        elif callable(obj):
            # this includes types.Function, types.BuiltinFunctionType,
            # types.BuiltinMethodType, functools.partial, functools.singledispatch,
            # "class funclike" from Lib/test/test_inspect... on and on it goes.
            ann = getattr(obj, "__annotations__", None)
            obj_globals = getattr(obj, "__globals__", None)
            obj_locals = None
            unwrap = obj
        else:
            raise TypeError(f"{obj!r} is not a module, class, or callable.")

        if ann is None:
            return {}

        if not isinstance(ann, dict):
            raise ValueError(f"{obj!r}.__annotations__ is neither a dict nor None")

        if not ann:
            return {}

        if not eval_str:
            return dict(ann)

        if unwrap is not None:
            while True:
                if hasattr(unwrap, "__wrapped__"):
                    unwrap = unwrap.__wrapped__  # type: ignore[reportFunctionMemberAccess,union-attr]
                    continue
                if isinstance(unwrap, functools.partial):
                    unwrap = unwrap.func  # type: ignore[reportGeneralTypeIssues,assignment]
                    continue
                break
            if hasattr(unwrap, "__globals__"):
                obj_globals = unwrap.__globals__  # type: ignore

        if globals is None:
            globals = obj_globals
        if locals is None:
            locals = obj_locals or {}

        # "Inject" type parameters into the local namespace
        # (unless they are shadowed by assignments *in* the local namespace),
        # as a way of emulating annotation scopes when calling `eval()`
        if type_params := getattr(obj, "__type_params__", ()):
            locals = {param.__name__: param for param in type_params} | locals

        return_value = {
            key: value if not isinstance(value, str) else eval(value, globals, locals)
            for key, value in ann.items()
        }
        return return_value


get_annotations = get_annotations



from __future__ import annotations

import inspect
from abc import ABC, abstractmethod
from collections.abc import Awaitable
from typing import Any, Callable, Generic, Optional

from nexusrpc._common import InputT, OutputT, ServiceHandlerT
from nexusrpc._service import Operation, OperationDefinition, ServiceDefinition
from nexusrpc._util import (
    get_operation,
    get_operation_factory,
    is_async_callable,
    is_callable,
)

from ._common import (
    CancelOperationContext,
    StartOperationContext,
    StartOperationResultAsync,
    StartOperationResultSync,
)


class OperationHandler(ABC, Generic[InputT, OutputT]):
    """
    Base class for an operation handler in a Nexus service implementation.

    To define a Nexus operation handler, create a method on your service handler class
    that takes `self` and returns an instance of :py:class:`OperationHandler`, and apply
    the :py:func:`@nexusrpc.handler.operation_handler` decorator.

    To create an operation handler that is limited to returning synchronously, use
    :py:func:`@nexusrpc.handler.SyncOperationHandler` to create the
    instance of :py:class:`OperationHandler` from the start method.
    """

    @abstractmethod
    def start(
        self, ctx: StartOperationContext, input: InputT
    ) -> (
        StartOperationResultSync[OutputT]
        | StartOperationResultAsync
        | Awaitable[StartOperationResultSync[OutputT] | StartOperationResultAsync]
    ):
        """
        Start the operation, completing either synchronously or asynchronously.

        Returns the result synchronously, or returns an operation token. Which path is
        taken may be decided at operation handling time.
        """
        ...

    @abstractmethod
    def cancel(self, ctx: CancelOperationContext, token: str) -> None | Awaitable[None]:
        """
        Cancel the operation.
        """
        ...


class SyncOperationHandler(OperationHandler[InputT, OutputT]):
    """
    An :py:class:`OperationHandler` that is limited to responding synchronously.

    This version of the class uses `async def` methods. For the syncio version, see
    :py:class:`nexusrpc.handler._syncio.SyncOperationHandler`.
    """

    def __init__(
        self, start: Callable[[StartOperationContext, InputT], Awaitable[OutputT]]
    ):
        if not is_async_callable(start):
            raise RuntimeError(
                f"{start} is not an `async def` method. "
                "SyncOperationHandler must be initialized with an `async def` method. "
            )
        self._start = start
        if start.__doc__:
            if start_func := getattr(self.start, "__func__", None):
                start_func.__doc__ = start.__doc__

    async def start(
        self, ctx: StartOperationContext, input: InputT
    ) -> StartOperationResultSync[OutputT]:
        """
        Start the operation and return its final result synchronously.

        The name 'SyncOperationHandler' means that it responds synchronously in the
        sense that the start method delivers the final operation result as its return
        value, rather than returning an operation token representing an in-progress
        operation. This version of the class uses `async def` methods. For the syncio
        version, see :py:class:`nexusrpc.handler.syncio.SyncOperationHandler`.
        """
        return StartOperationResultSync(await self._start(ctx, input))

    async def cancel(self, ctx: CancelOperationContext, token: str) -> None:
        raise NotImplementedError(
            "An operation that responded synchronously cannot be cancelled."
        )


class MiddlewareSafeOperationHandler(OperationHandler[Any, Any], ABC):
    """
    An :py:class:`OperationHandler` where :py:attr:`start` and :py:attr:`cancel`
    can be awaited by an async runtime. It can produce a result synchronously by returning
    :py:class:`StartOperationResultSync` or asynchronously by returning :py:class:`StartOperationResultAsync`
    in the same fashion that :py:class:`OperationHandler` does.

    .. warning::
        This API is experimental and may change in the future.
    """

    @abstractmethod
    async def start(
        self, ctx: StartOperationContext, input: Any
    ) -> StartOperationResultSync[Any] | StartOperationResultAsync:
        """
        Start the operation and return it's result or an async token.
        """
        ...

    @abstractmethod
    async def cancel(self, ctx: CancelOperationContext, token: str) -> None:
        """
        Cancel an in progress operation identified by the given token.
        """
        ...


def collect_operation_handler_factories_by_method_name(
    user_service_cls: type[ServiceHandlerT],
    service: Optional[ServiceDefinition],
) -> dict[str, Callable[[ServiceHandlerT], OperationHandler[Any, Any]]]:
    """
    Collect operation handler methods from a user service handler class.
    """
    factories: dict[str, Callable[[ServiceHandlerT], OperationHandler[Any, Any]]] = {}
    service_method_names = (
        {op_defn.method_name for op_defn in service.operation_definitions.values()}
        if service
        else set()
    )
    seen = set()
    for _, method in inspect.getmembers(user_service_cls, is_callable):
        if factory := get_operation_factory(method):
            if op := get_operation(factory):
                if op.name in seen:
                    raise RuntimeError(
                        f"Operation '{op.name}' in service '{user_service_cls.__name__}' "
                        f"is defined multiple times."
                    )
                if service and op.method_name not in service_method_names:
                    _names = ", ".join(f"'{s}'" for s in sorted(service_method_names))
                    msg = (
                        f"Operation method name '{op.method_name}' in service handler {user_service_cls} "
                        f"does not match an operation method name in the service definition. "
                        f"Available method names in the service definition: "
                    )
                    msg += _names if _names else "[none]"
                    raise TypeError(msg)

                assert op.method_name, (
                    f"Operation '{op}' method name should not be empty. Please report this as a bug."
                )
                factories[op.method_name] = factory
                seen.add(op.name)
    return factories


def validate_operation_handler_methods(
    service_cls: type[ServiceHandlerT],
    operation_handler_factories_by_method_name: dict[
        str, Callable[[ServiceHandlerT], OperationHandler[Any, Any]]
    ],
    service_definition: ServiceDefinition,
) -> None:
    """Validate operation handler methods against a service definition.

    For every operation in ``service_definition``:

    1. There must be a method in ``user_methods`` whose method name matches the method
       name from the service definition.

    2. The input and output types of the handler method must exactly match the types
       declared in the service definition.
    """
    operation_handler_factories_by_method_name = (
        operation_handler_factories_by_method_name.copy()
    )
    for op_defn in service_definition.operation_definitions.values():
        factory = operation_handler_factories_by_method_name.pop(
            op_defn.method_name, None
        )
        if not factory:
            raise TypeError(
                f"Service '{service_cls}' does not implement an operation with "
                f"method name '{op_defn.method_name}'. But this operation is in service "
                f"definition '{service_definition}'."
            )
        op = get_operation(factory)
        if not isinstance(op, Operation):
            raise ValueError(
                f"Method '{factory}' in class '{service_cls.__name__}' "
                f"does not have a valid __nexus_operation__ attribute. "
                f"Did you forget to decorate the operation method with an operation handler decorator such as "
                f":py:func:`@nexusrpc.handler.operation_handler`?"
            )
        if op.name not in [op.method_name, op_defn.name]:
            raise TypeError(
                f"Operation '{op_defn.method_name}' in service '{service_cls}' "
                f"has name '{op.name}', but the name in the service definition "
                f"is '{op_defn.name}'. Operation handlers may not override the name of an operation "
                f"in the service definition."
            )
        # If handler's input_type is None (missing annotation), skip validation - the handler
        # relies on the service definition for type information. This supports handlers without
        # explicit type annotations when a service definition is provided.
        if op.input_type is not None and op_defn.input_type != op.input_type:
            raise TypeError(
                f"OperationHandler input type mismatch for '{service_cls}.{op_defn.method_name}': "
                f"expected {op_defn.input_type}, got {op.input_type}"
            )

        # If handler's output_type is None (missing annotation), skip validation - the handler
        # relies on the service definition for type information.
        if op.output_type is not None and op.output_type != op_defn.output_type:
            raise TypeError(
                f"OperationHandler output type mismatch for '{service_cls}.{op_defn.method_name}': "
                f"expected {op_defn.output_type}, got {op.output_type}"
            )
    if operation_handler_factories_by_method_name:
        raise ValueError(
            f"Service '{service_cls}' implements more operations than the interface '{service_definition}'. "
            f"Extra operations: {', '.join(sorted(operation_handler_factories_by_method_name.keys()))}."
        )


def service_definition_from_operation_handler_methods(
    service_name: str,
    user_methods: dict[str, Callable[[ServiceHandlerT], OperationHandler[Any, Any]]],
) -> ServiceDefinition:
    """
    Create a service definition from operation handler factory methods.

    In general, users should have access to, or define, a service definition, and validate
    their service handler against it by passing the service definition to the
    :py:func:`@nexusrpc.handler.service_handler` decorator. This function is used when
    that is not the case.
    """
    op_defns: dict[str, OperationDefinition[Any, Any]] = {}
    for name, method in user_methods.items():
        op = get_operation(method)
        if not isinstance(op, Operation):
            raise ValueError(
                f"In service '{service_name}', could not locate operation definition for "
                f"user operation handler method '{name}'. Did you forget to decorate the operation "
                f"method with an operation handler decorator such as "
                f"@nexusrpc.handler.sync_operation?"
            )
        op_defns[op.name] = OperationDefinition.from_operation(op)

    return ServiceDefinition(name=service_name, operation_definitions=op_defns)



from __future__ import annotations

from typing import (
    Callable,
)

from nexusrpc._common import InputT, OutputT
from nexusrpc._util import (
    is_async_callable,
)
from nexusrpc.handler._common import (
    CancelOperationContext,
    StartOperationContext,
    StartOperationResultSync,
)

from ._operation_handler import OperationHandler


class SyncOperationHandler(OperationHandler[InputT, OutputT]):
    """
    An :py:class:`nexusrpc.handler.OperationHandler` that is limited to responding synchronously.

    The name 'SyncOperationHandler' means that it responds synchronously, in the
    sense that the start method delivers the final operation result as its return
    value, rather than returning an operation token representing an in-progress
    operation.

    This version of the class uses `def` methods. For the async version, see
    :py:class:`nexusrpc.handler.SyncOperationHandler`.
    """

    def __init__(self, start: Callable[[StartOperationContext, InputT], OutputT]):
        if is_async_callable(start):
            raise RuntimeError(
                f"{start} is an `async def` method. "
                "SyncOperationHandler must be initialized with a `def` method. "
                "To use `async def` methods, use nexusrpc.handler.SyncOperationHandler."
            )
        self._start = start
        if start.__doc__:
            if start_func := getattr(self.start, "__func__", None):
                start_func.__doc__ = start.__doc__

    def start(
        self, ctx: StartOperationContext, input: InputT
    ) -> StartOperationResultSync[OutputT]:
        """
        Start the operation and return its final result synchronously.
        """
        return StartOperationResultSync(self._start(ctx, input))

    def cancel(self, ctx: CancelOperationContext, token: str) -> None:
        raise NotImplementedError(
            "An operation that responded synchronously cannot be cancelled."
        )



"""
Test that operation decorators result in operation factories that return the correct result.
"""

from dataclasses import dataclass
from typing import Any, Union, cast

import pytest

import nexusrpc
from nexusrpc import InputT, OutputT
from nexusrpc._util import get_service_definition, is_async_callable
from nexusrpc.handler import (
    CancelOperationContext,
    OperationHandler,
    StartOperationContext,
    StartOperationResultAsync,
    StartOperationResultSync,
    service_handler,
    sync_operation,
)
from nexusrpc.handler._decorators import operation_handler
from nexusrpc.handler._operation_handler import (
    collect_operation_handler_factories_by_method_name,
)
from tests.helpers import TestOperationTaskCancellation


@dataclass()
class _TestCase:
    Service: type[Any]
    expected_operation_factories: dict[str, Any]


class ManualOperationDefinition(_TestCase):
    @service_handler
    class Service:
        @operation_handler
        def operation(self) -> OperationHandler[int, int]:
            class OpHandler(OperationHandler[int, int]):
                async def start(
                    self, ctx: StartOperationContext, input: int
                ) -> StartOperationResultSync[int]:
                    return StartOperationResultSync(7)

                def cancel(self, ctx: CancelOperationContext, token: str) -> None:
                    raise NotImplementedError

            return OpHandler()

    expected_operation_factories = {"operation": 7}


class SyncOperation(_TestCase):
    @service_handler
    class Service:
        @sync_operation
        async def sync_operation_handler(
            self, _ctx: StartOperationContext, _input: int
        ) -> int:
            return 7

    expected_operation_factories = {"sync_operation_handler": 7}  # type: ignore


@pytest.mark.parametrize(
    "test_case",
    [
        ManualOperationDefinition,
        SyncOperation,
    ],
)
@pytest.mark.asyncio
async def test_collected_operation_factories_match_service_definition(
    test_case: type[_TestCase],
):
    service = get_service_definition(test_case.Service)
    assert isinstance(service, nexusrpc.ServiceDefinition)
    assert service.name == "Service"
    operation_factories = collect_operation_handler_factories_by_method_name(
        test_case.Service, service
    )
    assert operation_factories.keys() == test_case.expected_operation_factories.keys()
    ctx = StartOperationContext(
        service="Service",
        operation="operation",
        headers={},
        request_id="request_id",
        task_cancellation=TestOperationTaskCancellation(),
    )

    async def execute(
        op: OperationHandler[InputT, OutputT],
        ctx: StartOperationContext,
        input: InputT,
    ) -> Union[
        StartOperationResultSync[OutputT],
        StartOperationResultAsync,
    ]:
        if is_async_callable(op.start):
            return await op.start(ctx, input)
        else:
            return cast(
                StartOperationResultSync[OutputT],
                op.start(ctx, input),
            )

    for op_name, expected_result in test_case.expected_operation_factories.items():
        op_factory = operation_factories[op_name]
        op = op_factory(test_case.Service)
        result = await execute(op, ctx, 0)
        assert isinstance(result, StartOperationResultSync)
        assert result.value == expected_result



from functools import partial

from nexusrpc._util import is_async_callable, is_callable


def test_def():
    def f(_a: int, _b: int) -> None:
        pass

    g = partial(f, _a=1)
    assert is_callable(f)
    assert is_callable(g)
    assert not is_async_callable(f)
    assert not is_async_callable(g)


def test_callable_instance():
    class f_cls:
        def __call__(self, _a: int, _b: int) -> None:
            pass

    f = f_cls()
    g = partial(f, _a=1)

    assert is_callable(f)
    assert is_callable(g)
    assert not is_async_callable(f)
    assert not is_async_callable(g)


def test_async_def():
    async def f(_a: int, _b: int) -> None:
        pass

    g = partial(f, _a=1)
    assert is_callable(f)
    assert is_callable(g)
    assert is_async_callable(f)
    assert is_async_callable(g)


def test_async_callable_instance():
    class f_cls:
        async def __call__(self, _a: int, _b: int) -> None:
            pass

    f = f_cls()
    g = partial(f, _a=1)

    assert is_callable(f)
    assert is_callable(g)
    assert is_async_callable(f)
    assert is_async_callable(g)


def test_partial():
    def f(_a: int, _b: int) -> None:
        pass

    _g = partial(f, _a=1)



from __future__ import annotations

import typing
import warnings
from collections.abc import Awaitable
from typing import (
    TYPE_CHECKING,
    Callable,
    Optional,
    TypeVar,
    Union,
)

from nexusrpc.handler import StartOperationContext

if TYPE_CHECKING:
    from nexusrpc import InputT, OutputT


ServiceHandlerT = TypeVar("ServiceHandlerT")


def get_start_method_input_and_output_type_annotations(
    start: Callable[
        [ServiceHandlerT, StartOperationContext, InputT],
        Union[OutputT, Awaitable[OutputT]],
    ],
) -> tuple[
    Optional[type[InputT]],
    Optional[type[OutputT]],
]:
    """Extract input and output type annotations from a start method.

    Args:
        start: A start method with signature (self, ctx: StartOperationContext, input: I) -> O

    Returns:
        A tuple of (input_type, output_type) where:

        - ``None`` means the type annotation is missing or could not be extracted.
          This is valid when a service definition provides the types.
        - ``type(None)`` (NoneType) means the annotation explicitly specified ``None``
          as the type (e.g., ``input: None`` or ``-> None``).

        When ``None`` is returned for either type, the caller should handle it based on
        whether a service definition is available:

        - If a service definition is provided, the types from the service definition
          will be used and type validation against the handler is skipped.
        - If no service definition is provided, an error will be raised downstream
          when attempting to create an ``OperationDefinition``.
    """
    try:
        type_annotations = typing.get_type_hints(start)
    except TypeError:
        return None, None
    output_type = type_annotations.pop("return", None)

    if len(type_annotations) != 2:
        input_type = None
    else:
        ctx_type, input_type = type_annotations.values()
        if not issubclass(ctx_type, StartOperationContext):
            warnings.warn(
                f"Expected first parameter of {start} to be an instance of "
                f"StartOperationContext, but is {ctx_type}."
            )
            input_type = None

    return input_type, output_type



from nexusrpc.handler import StartOperationContext, service_handler, sync_operation


# This file is for type-level testing only
def _():
    @service_handler
    class TestServiceHandler:
        @sync_operation
        def good_syncio(self, _ctx: StartOperationContext, input: int) -> str:
            return str(input)

        @sync_operation
        async def good_asyncio(self, _ctx: StartOperationContext, input: int) -> str:
            return str(input)

        @sync_operation
        async def another_asyncio_1(
            self, _ctx: StartOperationContext, input: int
        ) -> str:
            return await self.good_asyncio(_ctx, input)

        @sync_operation(name="custom_name")
        def good_syncio_with_name(self, _ctx: StartOperationContext, input: str) -> int:
            return len(input)

        @sync_operation(name="async_custom")
        async def good_asyncio_with_name(
            self, _ctx: StartOperationContext, input: str
        ) -> int:
            return len(input)

        @sync_operation
        async def another_asyncio_2(
            self, _ctx: StartOperationContext, input: str
        ) -> int:
            return await self.good_asyncio_with_name(_ctx, input)

        # assert-type-error-pyright: 'Argument of type .+ cannot be assigned to parameter'
        # assert-type-error-mypy: "has incompatible type"
        @sync_operation  # type: ignore
        def syncio_bad_signature_1(self, x: int) -> str:
            return str(x)

        # assert-type-error-pyright: 'Argument of type .+ cannot be assigned to parameter'
        # assert-type-error-mypy: "has incompatible type"
        @sync_operation  # type: ignore
        def syncio_bad_signature_2(self, x: int, _y: str, _z: float) -> str:
            return str(x)

        # assert-type-error-pyright: 'Argument of type .+ cannot be assigned to parameter'
        # assert-type-error-mypy: "has incompatible type"
        @sync_operation  # type: ignore
        def syncio_bad_signature_3(self) -> str:
            return "test"

        # assert-type-error-pyright: 'Argument of type .+ cannot be assigned to parameter'
        # assert-type-error-mypy: "has incompatible type"
        @sync_operation  # type: ignore
        async def asyncio_bad_signature_1(self, x: int) -> str:
            return str(x)

        # assert-type-error-pyright: 'Argument of type .+ cannot be assigned to parameter'
        # assert-type-error-mypy: "has incompatible type"
        @sync_operation(name="bad")  # type: ignore
        def syncio_bad_signature_with_name(self, x: int) -> str:
            return str(x)

        # assert-type-error-pyright: 'Argument of type .+ cannot be assigned to parameter'
        # assert-type-error-mypy: "has incompatible type"
        @sync_operation(name="bad")  # type: ignore
        async def asyncio_bad_signature_with_name(self, x: int) -> str:
            return str(x)

    _ = TestServiceHandler



# It's important to test that retrieving annotation works both with and without `from __future__ import annotations`.
# Currently we do so by applying it in some test files and not others.
# See https://docs.python.org/3/howto/annotations.html#accessing-the-annotations-dict-of-an-object-in-python-3-9-and-older
from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Optional

import pytest

import nexusrpc
from nexusrpc import Operation, ServiceDefinition
from nexusrpc._util import get_service_definition

# See https://docs.python.org/3/howto/annotations.html


@dataclass()
class _TestCase:
    UserService: type[Any]
    expected_operation_names: set[str]
    expected_error: Optional[str] = None


class TypeAnnotationsOnly(_TestCase):
    @nexusrpc.service
    class A1:
        a: Operation[int, str]

    # A2 intentionally omits @nexusrpc.service at definition time.
    # The decorator will be applied in the test function to verify that
    # service definitions correctly inherit operations from decorated parent classes.
    class A2(A1):
        b: Operation[int, str]  # type: ignore[reportUninitializedInstanceVariable]

    UserService = A2
    expected_operation_names = {"a", "b"}


class TypeAnnotationsWithValues(_TestCase):
    @nexusrpc.service
    class A1:
        a: Operation[int, str] = Operation[int, str](name="a-name")

    # A2 intentionally omits @nexusrpc.service at definition time.
    # The decorator will be applied in the test function to verify that
    # service definitions correctly inherit operations from decorated parent classes.
    class A2(A1):
        b: Operation[int, str] = Operation[int, str](name="b-name")

    UserService = A2
    expected_operation_names = {"a-name", "b-name"}


class TypeAnnotationsWithValuesAllFromParentClass(_TestCase):
    # See https://docs.python.org/3/howto/annotations.html#accessing-the-annotations-dict-of-an-object-in-python-3-9-and-older
    # A2.__annotations__ returns annotations from parent
    @nexusrpc.service
    class A1:
        a: Operation[int, str] = Operation[int, str](name="a-name")
        b: Operation[int, str] = Operation[int, str](name="b-name")

    class A2(A1):
        pass

    UserService = A2
    expected_operation_names = {"a-name", "b-name"}


class TypeAnnotationWithInheritedInstance(_TestCase):
    @nexusrpc.service
    class A1:
        a: Operation[int, str] = Operation[int, str](name="a-name")

    class A2(A1):
        a: Operation[int, str]

    UserService = A2
    expected_operation_names = {"a-name", "b-name"}


class InstanceWithoutTypeAnnotationIsAnError(_TestCase):
    class A1:
        a = Operation[int, str](name="a-name")

    UserService = A1
    expected_error = "Operation 'a-name' has no input type"


class InvalidUseOfTypeAsValue(_TestCase):
    class A1:
        a = Operation[int, str]

    UserService = A1
    expected_error = "Did you accidentally use '=' instead of ':'"


class ChildClassSynthesizedWithTypeValues(_TestCase):
    @nexusrpc.service
    class A1:
        a: Operation[int, str]

    A2 = type("A2", (A1,), {name: Operation[int, str] for name in ["b"]})

    UserService = A2
    expected_error = "Did you accidentally use '=' instead of ':'"


@pytest.mark.parametrize(
    "test_case",
    [
        TypeAnnotationsOnly,
        TypeAnnotationsWithValues,
        TypeAnnotationsWithValuesAllFromParentClass,
        InstanceWithoutTypeAnnotationIsAnError,
        InvalidUseOfTypeAsValue,
        ChildClassSynthesizedWithTypeValues,
    ],
)
def test_user_service_definition_inheritance(test_case: type[_TestCase]):
    if test_case.expected_error:
        with pytest.raises(Exception, match=test_case.expected_error):
            nexusrpc.service(test_case.UserService)
        return

    service_defn = get_service_definition(nexusrpc.service(test_case.UserService))
    assert isinstance(service_defn, ServiceDefinition)
    assert set(service_defn.operation_definitions) == test_case.expected_operation_names
    for op in service_defn.operation_definitions.values():
        assert op.input_type is int
        assert op.output_type is str



from __future__ import annotations

from collections.abc import AsyncIterable, Awaitable, Mapping
from dataclasses import dataclass
from typing import (
    Any,
    Optional,
    Protocol,
    Union,
)


@dataclass(frozen=True)
class Content:
    """
    A container for a map of headers and a byte array of data.

    It is used by the SDK's Serializer interface implementations.

    .. warning::
        This API is experimental and may change in the future.
    """

    headers: Mapping[str, str]
    """
    Header that should include information on how to deserialize this content.
    Headers constructed by the framework always have lower case keys.
    User provided keys are treated case-insensitively.
    """

    data: bytes
    """Request or response data."""


class Serializer(Protocol):
    """
    Serializer is used by the framework to serialize/deserialize input and output.

    .. warning::
        This API is experimental and may change in the future.
    """

    def serialize(self, value: Any) -> Union[Content, Awaitable[Content]]:
        """Serialize encodes a value into a Content."""
        ...

    def deserialize(
        self, content: Content, as_type: Optional[type[Any]] = None
    ) -> Union[Any, Awaitable[Any]]:
        """Deserialize decodes a Content into a value.

        Args:
            content: The content to deserialize.
            as_type: The type to convert the result of deserialization into.
                     Do not attempt type conversion if this is None.
        """
        ...


class LazyValueT(Protocol):
    def consume(
        self, as_type: Optional[type[Any]] = None
    ) -> Union[Any, Awaitable[Any]]: ...


class LazyValue(LazyValueT):
    """
    A container for a value encoded in an underlying stream.

    It is used to stream inputs and outputs in the various client and server APIs.

    .. warning::
        This API is experimental and may change in the future.

    Example:
        .. code-block:: python

            # Creating a LazyValue from raw data
            lazy_input = LazyValue(
                serializer=my_serializer,
                headers={"content-type": "application/json"},
                stream=async_data_stream
            )

            # Using LazyValue with Handler.start_operation
            handler = nexusrpc.handler.Handler([my_service])
            result = await handler.start_operation(ctx, lazy_input)

            # Consuming a LazyValue directly
            deserialized_data = await lazy_input.consume(as_type=MyInputType)
    """

    def __init__(
        self,
        serializer: Serializer,
        headers: Mapping[str, str],
        stream: Optional[AsyncIterable[bytes]] = None,
    ) -> None:
        """
        Args:
            serializer: The serializer to use for consuming the value.
            headers: Headers that include information on how to process the stream's content.
                     Headers constructed by the framework always have lower case keys.
                     User provided keys are treated case-insensitively.
            stream:  AsyncIterable of bytes that contains request or response data.
                     None means empty data.
        """
        self.serializer = serializer
        self.headers = headers
        self.stream = stream

    async def consume(self, as_type: Optional[type[Any]] = None) -> Any:
        """
        Consume the underlying reader stream, deserializing via the embedded serializer.
        """
        if self.stream is None:
            return await self.serializer.deserialize(
                Content(headers=self.headers, data=b""), as_type=as_type
            )

        return await self.serializer.deserialize(
            Content(
                headers=self.headers,
                data=b"".join([c async for c in self.stream]),
            ),
            as_type=as_type,
        )



"""
Tests for invalid ways that users may attempt to write service definition and service
handler implementations.
"""

from dataclasses import dataclass

import pytest

import nexusrpc
from nexusrpc.handler import (
    Handler,
    StartOperationContext,
    service_handler,
    sync_operation,
)
from nexusrpc.handler._decorators import operation_handler
from nexusrpc.handler._operation_handler import OperationHandler


@dataclass()
class _TestCase:
    error_message: str

    @staticmethod
    def build():
        pass


class OperationHandlerOverridesNameInconsistentlyWithServiceDefinition(_TestCase):
    @staticmethod
    def build():
        @nexusrpc.service
        class SD:
            my_op: nexusrpc.Operation[None, None]

        @service_handler(service=SD)
        class SH:
            @sync_operation(name="foo")
            async def my_op(
                self, _ctx: StartOperationContext, _input: None
            ) -> None: ...

        _ = SH

    error_message = "Operation handlers may not override the name of an operation in the service definition"


class ServiceDefinitionHasExtraOp(_TestCase):
    @staticmethod
    def build():
        @nexusrpc.service
        class SD:
            my_op_1: nexusrpc.Operation[None, None]
            my_op_2: nexusrpc.Operation[None, None]

        @service_handler(service=SD)
        class SH:
            @sync_operation
            async def my_op_1(
                self, _ctx: StartOperationContext, _input: None
            ) -> None: ...

        _ = SH

    error_message = "does not implement an operation with method name 'my_op_2'"


class ServiceHandlerHasExtraOp(_TestCase):
    @staticmethod
    def build():
        @nexusrpc.service
        class SD:
            my_op_1: nexusrpc.Operation[None, None]

        @service_handler(service=SD)
        class SH:
            @sync_operation
            async def my_op_1(
                self, _ctx: StartOperationContext, _input: None
            ) -> None: ...

            @sync_operation
            async def my_op_2(
                self, _ctx: StartOperationContext, _input: None
            ) -> None: ...

        _ = SH

    error_message = "does not match an operation method name in the service definition"


class ServiceDefinitionOperationHasNoTypeParams(_TestCase):
    @staticmethod
    def build():
        @nexusrpc.service
        class SD:
            my_op: nexusrpc.Operation  # type: ignore

        @service_handler(service=SD)
        class SH:
            @sync_operation
            async def my_op(
                self, _ctx: StartOperationContext, _input: None
            ) -> None: ...

        _ = SH

    error_message = "has 0 type parameters"


class AsyncioHandlerWithSyncioOperation(_TestCase):
    @staticmethod
    def build():
        @service_handler
        class SH:
            @sync_operation
            def my_op(self, _ctx: StartOperationContext, _input: None) -> None: ...

        _ = Handler([SH()])

    error_message = "you have not supplied an executor"


class OperationHandlerNoInputOutputTypeAnnotationsWithoutServiceDefinition(_TestCase):
    @staticmethod
    def build():
        @service_handler
        class SubclassingNoInputOutputTypeAnnotationsWithoutServiceDefinition:
            @operation_handler
            def op(self) -> OperationHandler: ...  # type: ignore

        _ = SubclassingNoInputOutputTypeAnnotationsWithoutServiceDefinition

    error_message = r"has no input type"


class SyncOperationNoTypeAnnotationsWithoutServiceDefinition(_TestCase):
    """Test that @sync_operation with missing type annotations raises an error
    when no service definition is provided.

    This tests the flow through get_start_method_input_and_output_type_annotations()
    which returns (None, None) for missing types, followed by
    OperationDefinition.from_operation() which raises ValueError.
    """

    @staticmethod
    def build():
        @service_handler
        class ServiceWithMissingTypes:
            @sync_operation
            async def op(self, ctx, input): ...  # type: ignore[reportMissingParameterType]

        _ = ServiceWithMissingTypes

    error_message = r"has no input type"


@pytest.mark.parametrize(
    "test_case",
    [
        OperationHandlerOverridesNameInconsistentlyWithServiceDefinition,
        ServiceDefinitionOperationHasNoTypeParams,
        ServiceDefinitionHasExtraOp,
        ServiceHandlerHasExtraOp,
        AsyncioHandlerWithSyncioOperation,
        OperationHandlerNoInputOutputTypeAnnotationsWithoutServiceDefinition,
        SyncOperationNoTypeAnnotationsWithoutServiceDefinition,
    ],
)
def test_invalid_usage(test_case: _TestCase):
    if test_case.error_message:
        with pytest.raises(Exception, match=test_case.error_message):
            test_case.build()
    else:
        test_case.build()



"""
nexusrpc is a library for building Nexus handlers.

See https://github.com/nexus-rpc and https://github.com/nexus-rpc/api/blob/main/SPEC.md.

Nexus is a synchronous RPC protocol. Arbitrary duration operations are modeled on top of
a set of pre-defined synchronous RPCs.

A Nexus caller calls a handler. The handler may respond inline (synchronous response) or
return a token referencing the ongoing operation (asynchronous response). The caller can
cancel an asynchronous operation, check for its outcome, or fetch its current state. The
caller can also specify a callback URL, which the handler uses to deliver the result of
an asynchronous operation when it is ready.
"""

from __future__ import annotations

from . import handler
from ._common import (
    Failure,
    HandlerError,
    HandlerErrorType,
    InputT,
    Link,
    OperationError,
    OperationErrorState,
    OutputT,
)
from ._serializer import Content, LazyValue, LazyValueT, Serializer
from ._service import Operation, OperationDefinition, ServiceDefinition, service
from ._util import (
    get_operation,
    get_service_definition,
    set_operation,
)

__all__ = [
    "Content",
    "Failure",
    "get_operation",
    "get_service_definition",
    "handler",
    "HandlerError",
    "HandlerErrorType",
    "InputT",
    "LazyValue",
    "LazyValueT",
    "Link",
    "Operation",
    "OperationDefinition",
    "OperationError",
    "OperationErrorState",
    "OutputT",
    "Serializer",
    "service",
    "ServiceDefinition",
    "set_operation",
]



import asyncio

import pytest

from nexusrpc import LazyValue
from nexusrpc.handler import (
    CancelOperationContext,
    Handler,
    OperationHandler,
    StartOperationContext,
    StartOperationResultAsync,
    StartOperationResultSync,
    operation_handler,
    service_handler,
    sync_operation,
)
from tests.helpers import DummySerializer, TestOperationTaskCancellation


class CancellableAsyncOperationHandler(OperationHandler[None, None]):
    async def start(
        self, ctx: StartOperationContext, input: None
    ) -> StartOperationResultAsync:
        try:
            await asyncio.wait_for(
                ctx.task_cancellation.wait_until_cancelled(), timeout=1
            )
        except TimeoutError as err:
            raise RuntimeError("Expected cancellation") from err

        details = ctx.task_cancellation.cancellation_reason()
        if not details:
            raise RuntimeError("Expected cancellation details")

        # normally you return a token but for this test
        # we use the token to indicate success by returning the expected
        # cancellation details
        return StartOperationResultAsync(details)

    async def cancel(self, ctx: CancelOperationContext, token: str) -> None:
        pass


@service_handler
class MyService:
    @operation_handler
    def cancellable_async(self) -> OperationHandler[None, None]:
        return CancellableAsyncOperationHandler()

    @sync_operation
    async def cancellable_sync(self, ctx: StartOperationContext, _input: None) -> str:
        cancelled = ctx.task_cancellation.wait_until_cancelled_sync(1)
        if not cancelled:
            raise RuntimeError("Expected cancellation")

        details = ctx.task_cancellation.cancellation_reason()
        if not details:
            raise RuntimeError("Expected cancellation details")

        return details


@pytest.mark.asyncio
async def test_cancellation_sync_operation():
    handler = Handler(user_service_handlers=[MyService()])
    cancellation = TestOperationTaskCancellation()
    start_ctx = StartOperationContext(
        service="MyService",
        operation="cancellable_sync",
        headers={},
        request_id="request_id",
        task_cancellation=cancellation,
    )

    operation_task = asyncio.create_task(
        handler.start_operation(
            start_ctx, LazyValue(serializer=DummySerializer(None), headers={})
        )
    )

    cancellation.cancel()
    result = await operation_task
    assert result == StartOperationResultSync("test cancellation occurred")


@pytest.mark.asyncio
async def test_cancellation_async_operation():
    handler = Handler(user_service_handlers=[MyService()])
    cancellation = TestOperationTaskCancellation()
    start_ctx = StartOperationContext(
        service="MyService",
        operation="cancellable_async",
        headers={},
        request_id="request_id",
        task_cancellation=cancellation,
    )

    operation_task = asyncio.create_task(
        handler.start_operation(
            start_ctx, LazyValue(serializer=DummySerializer(None), headers={})
        )
    )

    cancellation.cancel()
    result = await operation_task
    assert result == StartOperationResultAsync("test cancellation occurred")



from dataclasses import dataclass

import pytest

import nexusrpc
from nexusrpc._util import get_service_definition


@dataclass
class NameOverrideTestCase:
    Interface: type
    expected_name: str


class NotCalled(NameOverrideTestCase):
    @nexusrpc.service
    class Interface:
        pass

    expected_name = "Interface"


class CalledWithoutArgs(NameOverrideTestCase):
    @nexusrpc.service()
    class Interface:
        pass

    expected_name = "Interface"


class CalledWithNameArg(NameOverrideTestCase):
    @nexusrpc.service(name="my-service-interface-🌈")
    class Interface:
        pass

    expected_name = "my-service-interface-🌈"


@pytest.mark.parametrize(
    "test_case",
    [
        NotCalled,
        CalledWithoutArgs,
        CalledWithNameArg,
    ],
)
def test_interface_name_overrides(test_case: type[NameOverrideTestCase]):
    defn = get_service_definition(test_case.Interface)
    assert defn
    assert defn.name == test_case.expected_name


def test_name_must_not_be_empty():
    with pytest.raises(ValueError):
        _ = nexusrpc.service(name="")(object)



"""
This file contains a test allowing assertions to be made that an expected type error is in
fact produced by the type-checker. I.e. that the type checker is not delivering a false
negative.

To use the test, add a comment of the following form to your test code:

    # assert-type-error-pyright: 'No overloads for "execute_operation" match' await
    nexus_client.execute_operation(  # type: ignore

The `type: ignore` is only necessary if your test code is being type-checked.
"""

import itertools
import json
import os
import platform
import re
import subprocess
import tempfile
from pathlib import Path

import pytest


def pytest_generate_tests(metafunc: pytest.Metafunc) -> None:
    """Dynamically generate test cases for files with type error assertions."""
    if metafunc.function.__name__ in [
        "test_type_errors_pyright",
        "test_type_errors_mypy",
    ]:
        tests_dir = Path(__file__).parent
        files_with_assertions = []

        for test_file in tests_dir.rglob("test_*.py"):
            if test_file.name == "test_type_errors.py":
                continue

            if _has_type_error_assertions(test_file):
                files_with_assertions.append(test_file)

        metafunc.parametrize("test_file", files_with_assertions, ids=lambda f: f.name)


@pytest.mark.skipif(platform.system() == "Windows", reason="TODO: broken on Windows")
def test_type_errors_pyright(test_file: Path):
    """
    Validate type error assertions in a single test file using pyright.

    For each line with a comment of the form `# assert-type-error-pyright: "regex"`,
    verify that pyright reports an error on the next non-comment line matching the regex.
    Also verify that there are no unexpected type errors.
    """
    _test_type_errors(
        test_file,
        _get_expected_errors(test_file, "pyright"),
        _get_pyright_errors(test_file),
    )


# This test is disabled since we currently have no way to be able to
def _test_type_errors(
    test_file: Path,
    expected_errors: dict[int, str],
    actual_errors: dict[int, str],
) -> None:
    for line_num, expected_pattern in expected_errors.items():
        if line_num not in actual_errors:
            pytest.fail(
                f"{test_file}:{line_num}: Expected type error matching '{expected_pattern}' but no error found"
            )

        actual_msg = actual_errors[line_num]
        if not re.search(expected_pattern, actual_msg):
            pytest.fail(
                f"{test_file}:{line_num}: Expected error matching '{expected_pattern}' but got '{actual_msg}'"
            )

    for line_num, actual_msg in actual_errors.items():
        if line_num not in expected_errors:
            pytest.fail(f"{test_file}:{line_num}: Unexpected type error: {actual_msg}")


def _has_type_error_assertions(test_file: Path) -> bool:
    """Check if a file contains any type error assertions."""
    with open(test_file) as f:
        return any(
            re.search(r"# assert-type-error-\w+:", line) for line in f.readlines()
        )


def _get_expected_errors(test_file: Path, type_checker: str) -> dict[int, str]:
    """Parse expected type errors from comments in a file for the specified type checker."""
    expected_errors = {}

    with open(test_file) as f:
        lines = zip(itertools.count(1), f)
        for line_num, line in lines:
            if match := re.search(
                rf'# assert-type-error-{re.escape(type_checker)}:\s*["\'](.+)["\']',
                line,
            ):
                pattern = match.group(1)
                for line_num, line in lines:
                    if line.strip() and not line.strip().startswith("#"):
                        expected_errors[line_num] = pattern
                        break

    return expected_errors


def _get_pyright_errors(test_file: Path) -> dict[int, str]:
    """Run pyright on a file and parse the actual type errors."""
    with tempfile.NamedTemporaryFile(mode="w", suffix=".json", delete=False) as f:
        # Create a temporary config file to disable type ignore comments
        config_data = {"enableTypeIgnoreComments": False}
        json.dump(config_data, f)
        config_path = f.name

    try:
        result = subprocess.run(
            ["uv", "run", "pyright", "--project", config_path, str(test_file)],
            capture_output=True,
            text=True,
        )

        actual_errors = {}
        abs_path = test_file.resolve()

        for line in result.stdout.splitlines():
            # pyright output format: /full/path/to/file.py:line:column - error: message (error_code)
            if match := re.match(
                rf"\s*{re.escape(str(abs_path))}:(\d+):\d+\s*-\s*error:\s*(.+)", line
            ):
                line_num = int(match.group(1))
                error_msg = match.group(2).strip()
                # Remove error code in parentheses if present
                error_msg = re.sub(r"\s*\([^)]+\)$", "", error_msg)
                actual_errors[line_num] = error_msg

        return actual_errors
    finally:
        if os.path.exists(config_path):
            os.unlink(config_path)


def _get_mypy_errors(test_file: Path) -> dict[int, str]:  # pyright: ignore[reportUnusedFunction]
    """Run mypy on a file and parse the actual type errors.

    Note: mypy does not have a direct equivalent to pyright's enableTypeIgnoreComments=false,
    so type ignore comments will still be respected by mypy. Users should avoid placing
    # type: ignore comments on lines they want to test, or manually remove them for testing.
    """
    result = subprocess.run(
        ["uv", "run", "mypy", str(test_file)],
        capture_output=True,
        text=True,
    )

    actual_errors = {}
    abs_path = test_file.resolve()

    for line in result.stdout.splitlines():
        # mypy output format: file.py:line: error: message
        if match := re.match(
            rf"{re.escape(str(abs_path))}:(\d+):\s*error:\s*(.+)", line
        ):
            line_num = int(match.group(1))
            error_msg = match.group(2).strip()
            actual_errors[line_num] = error_msg

    return actual_errors



import asyncio
import threading
from dataclasses import dataclass
from typing import Any, Optional

from nexusrpc import Content
from nexusrpc.handler import OperationTaskCancellation


@dataclass
class DummySerializer:
    value: Any

    async def serialize(self, value: Any) -> Content:  # pyright: ignore[reportUnusedParameter]
        raise NotImplementedError

    async def deserialize(
        self,
        content: Content,  # pyright: ignore[reportUnusedParameter]
        as_type: Optional[type[Any]] = None,  # pyright: ignore[reportUnusedParameter]
    ) -> Any:
        return self.value


class TestOperationTaskCancellation(OperationTaskCancellation):
    __test__ = False

    # A naive implementation of cancellation for use in tests
    def __init__(self):
        self._details = None
        self._evt = threading.Event()
        self._lock = threading.Lock()

    def is_cancelled(self) -> bool:
        return self._evt.is_set()

    def cancellation_reason(self) -> Optional[str]:
        with self._lock:
            return self._details

    def wait_until_cancelled_sync(self, timeout: float | None = None) -> bool:
        return self._evt.wait(timeout)

    async def wait_until_cancelled(self):
        while not self.is_cancelled():
            await asyncio.sleep(0.05)

    def cancel(self):
        with self._lock:
            self._details = "test cancellation occurred"
            self._evt.set()
